"""Implements regression Metrics."""
import torch

from torchflare.metrics.meters import MetricMeter


def _check_shape(outputs: torch.Tensor, targets: torch.Tensor):
    """Check that predictions and target have the same shape, else raise error.

    Raises:
        ValueError: If shapes of outputs and targets is not same.

    Args:
        outputs: torch.Tensor
        targets: torch.Tensor
    """
    if outputs.shape != targets.shape:
        raise ValueError("Predictions and targets are expected to have the same shape")


def detach_tensor(x):
    """Detaches tensor from the graph and moves to cpu.

    Args:
        x:torch.Tensor

    Returns:
        torch.Tensor
    """
    return x.detach().cpu()


class MAE(MetricMeter):
    """Computes Mean Absolute Error."""

    def __init__(self):
        """Constructor method for MAE."""
        self._n_obs = None
        self._abs_error_sum = None

        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The name of the class
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs, targets):
        """Accumulates the batch outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : targets to use for computing accuracy
        """
        outputs, targets = detach_tensor(outputs), detach_tensor(targets)
        _check_shape(outputs, targets)
        self._abs_error_sum += torch.sum(torch.abs(outputs - targets))
        self._n_obs += targets.numel()

    def reset(self):
        """Reset the output and target lists."""
        self._n_obs = torch.tensor(0)
        self._abs_error_sum = torch.tensor(0.0)

    def compute(self) -> torch.Tensor:
        """Computes the MAE.

        Returns:
            The computed MAE.
        """
        return self._abs_error_sum / self._n_obs


class MSE(MetricMeter):
    """Computes Mean Squared Error."""

    def __init__(self):
        """Constructor Method for MSE."""
        self._n_obs = None
        self._squared_error_sum = None

        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The name of the class
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs, targets):
        """Accumulates the batch outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : targets to use for computing accuracy
        """
        outputs, targets = detach_tensor(outputs), detach_tensor(targets)
        _check_shape(outputs, targets)
        self._squared_error_sum += torch.sum(torch.pow(outputs - targets, 2))
        self._n_obs += targets.numel()

    def reset(self):
        """Reset the output and target lists."""
        self._n_obs = torch.tensor(0)
        self._squared_error_sum = torch.tensor(0.0)

    def compute(self) -> torch.Tensor:
        """Computes the MSE.

        Returns:
            The computed MSE.
        """
        return self._squared_error_sum / self._n_obs


class MSLE(MetricMeter):
    """Computes Mean Squared Log Error."""

    def __init__(self):
        """Constructor Method for MSLE."""
        self._n_obs = None
        self._log_squared_error_sum = None
        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The name of the class
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs, targets):
        """Accumulates the batch outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : targets to use for computing accuracy
        """
        outputs, targets = detach_tensor(outputs), detach_tensor(targets)
        _check_shape(outputs, targets)
        diff = torch.log1p(outputs) - torch.log1p(targets)
        self._log_squared_error_sum += torch.sum(torch.pow(diff, 2))
        self._n_obs += targets.numel()

    def reset(self):
        """Reset the output and target lists."""
        self._n_obs = torch.tensor(0)
        self._log_squared_error_sum = torch.tensor(0.0)

    def compute(self) -> torch.Tensor:
        """Computes the MSLE.

        Returns:
            The computed MSLE.
        """
        return self._log_squared_error_sum / self._n_obs


class R2Score(MetricMeter):
    """Computes R2-score."""

    def __init__(self):
        """Constructor method for R2-score."""
        self._num_examples = None
        self._sum_of_errors = None
        self._y_sq_sum = None
        self._y_sum = None

    def reset(self) -> None:
        """Reset the output and target lists."""
        self._num_examples = 0
        self._sum_of_errors = torch.tensor(0.0)
        self._y_sq_sum = torch.tensor(0.0)
        self._y_sum = torch.tensor(0.0)

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The name of the class
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs, targets):
        """Accumulates the batch outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : targets to use for computing accuracy
        """
        self._num_examples += outputs.shape[0]
        self._sum_of_errors += torch.sum(torch.pow(outputs - targets, 2))

        self._y_sum += torch.sum(targets)
        self._y_sq_sum += torch.sum(torch.pow(targets, 2))

    def compute(self) -> torch.Tensor:
        """Computes the R2Score.

        Raises:
            ValueError:  If no examples are found.

        Returns:
            The computed R2Score.
        """
        if self._num_examples == 0:
            raise ValueError("R2Score must have at least one example before it can be computed.")
        return 1 - self._sum_of_errors / (self._y_sq_sum - (self._y_sum ** 2) / self._num_examples)


__all__ = ["MAE", "MSE", "MSLE", "R2Score"]



"""Wrapper for dataloaders."""
from __future__ import annotations

from typing import List, Optional, Tuple, Union

import albumentations as A
import pandas as pd
import torchvision
from torch.utils.data import DataLoader

from torchflare.datasets.classification import ImageDataset
from torchflare.datasets.segmentation import SegmentationDataset
from torchflare.datasets.tabular import TabularDataset
from torchflare.datasets.text_dataset import TextClassificationDataset


class SimpleDataloader:
    """Class to create easy to use dataloaders."""

    def __init__(self, ds):
        """Constructor method.

        Args:
            ds : A pytorch style dataset having __len__ and __getitem__ methods.
        """
        self.ds = ds

    @classmethod
    def image_data_from_df(
        cls,
        path: str,
        df: pd.DataFrame,
        image_col: str,
        label_cols: List[str] = None,
        augmentations: Optional[Union[A.Compose, torchvision.transforms.Compose]] = None,
        convert_mode: str = "RGB",
        extension: str = None,
    ) -> SimpleDataloader:
        """Classmethod to create a dataset for image data when you have image names/ids , labels in dataframe.

        Args:
            path: The path where images are saved.
            df: The dataframe containing the image name/ids, and the targets
            image_col: The name of the image column containing the image name/ids along with image extension.
                i.e. the images should have names like img_215.jpg or img_name.png ,etc
            augmentations: The batch_mixers to be used on images.
            label_cols: The list of columns containing targets.
            extension : The image file extension.
            convert_mode: The mode to be passed to PIL.Image.convert.

        Returns:
            Pytorch dataset created from dataframe.

        Note:

            For inference do not pass in the label_cols, keep it None.

            Augmentations must be Compose objects from albumentations or torchvision.

        """
        return cls(
            ImageDataset.from_df(
                path=path,
                df=df,
                image_col=image_col,
                label_cols=label_cols,
                augmentations=augmentations,
                convert_mode=convert_mode,
                extension=extension,
            )
        )

    @classmethod
    def image_data_from_folders(
        cls,
        path: str,
        augmentations: Optional[Union[A.Compose, torchvision.transforms.Compose]] = None,
        convert_mode: str = "RGB",
    ) -> SimpleDataloader:
        """Classmethod to create pytorch dataset from folders.

        Args:
            path: The path where images are stored.
            augmentations:The batch_mixers to be used on images.
            convert_mode: The mode to be passed to PIL.Image.convert.

        Returns:
            Pytorch image dataset created from folders

        Note:

            Augmentations must be Compose objects from albumentations or torchvision.

            The training directory structure should be as follows:
                train/class_1/xxx.jpg
                .
                .
                train/class_n/xxz.jpg

            The test directory structure should be as follows:
                test_dir/xxx.jpg
                test_dir/xyz.jpg
                test_dir/ppp.jpg

        """
        return cls(ImageDataset.from_folders(path=path, augmentations=augmentations, convert_mode=convert_mode))

    @classmethod
    def segmentation_data_from_rle(
        cls,
        path: str,
        df: pd.DataFrame,
        image_col: str,
        mask_cols: List[str] = None,
        augmentations: Optional[Union[A.Compose, torchvision.transforms.Compose]] = None,
        mask_size: Tuple[int, int] = None,
        num_classes: List = None,
        extension: Optional[str] = None,
        image_convert_mode: str = "RGB",
    ) -> SimpleDataloader:
        """Classmethod to create a dataset for segmentation data when you have, rule length encodings stored in a dataframe.

        Args:
            path: The path where images are saved.
            df: The dataframe containing the image name/ids, and the targets
            image_col: The name of the image column containing the image name/ids along with image extension.
                i.e. the images should have names like img_215.jpg or img_name.png ,etc
            augmentations: The batch_mixers to be used on images and the masks.
            mask_cols: The list of columns containing the rule length encoding.
            mask_size: The size of mask.
            num_classes: The list of num_classes.
            extension : The image file extension.
            image_convert_mode: The mode to be passed to PIL.Image.convert.

        Returns:
            Pytorch dataset created from Rule-length encodings

        Note:

            This method will make only binary masks.

            If you want to create a dataloader for testing set mask_cols = None, mask_size = None, num_classes = None.

        """
        return cls(
            SegmentationDataset.from_rle(
                path=path,
                df=df,
                image_col=image_col,
                mask_cols=mask_cols,
                augmentations=augmentations,
                mask_size=mask_size,
                num_classes=num_classes,
                image_convert_mode=image_convert_mode,
                extension=extension,
            )
        )

    @classmethod
    def segmentation_data_from_folders(
        cls,
        image_path: str,
        mask_path: str = None,
        augmentations: Optional[Union[A.Compose, torchvision.transforms.Compose]] = None,
        image_convert_mode: str = "L",
        mask_convert_mode: str = "L",
    ) -> SimpleDataloader:
        """Classmethod to create pytorch dataset from folders.

        Args:
            image_path: The path where images are stored.
            mask_path: The path where masks are stored.
            augmentations: The batch_mixers to apply on images and masks.
            image_convert_mode: The mode to be passed to PIL.Image.convert for input images
            mask_convert_mode: The mode to be passed to PIL.Image.convert for masks.

        Returns:
            Pytorch Segmentation dataset created from folders.
        """
        return cls(
            SegmentationDataset.from_folders(
                image_path=image_path,
                mask_path=mask_path,
                augmentations=augmentations,
                image_convert_mode=image_convert_mode,
                mask_convert_mode=mask_convert_mode,
            )
        )

    @classmethod
    def tabular_data_from_df(
        cls, df: pd.DataFrame, feature_cols: Union[str, List[str]], label_cols: Optional[Union[str, List[str]]] = None,
    ) -> SimpleDataloader:
        """Classmethod to create dataset for tabular data from dataframe.

        Args:
            df: The dataframe containing features and labels.
            feature_cols: name(str) or list containing names feature columns.
            label_cols: name(str) or list containing names label columns.

        Returns:
            Tabular pytorch dataset
        """
        return cls(TabularDataset.from_df(df=df, feature_cols=feature_cols, label_cols=label_cols))

    @classmethod
    def tabular_data_from_csv(
        cls, csv_path: str, feature_cols: Union[str, List[str]], label_cols: Optional[Union[str, List[str]]] = None,
    ) -> SimpleDataloader:
        """Classmethod to create a dataset for tabular data from csv.

        Args:
            csv_path: The full path to csv.
            feature_cols: name(str) or list containing names feature columns.
            label_cols: name(str) or list containing names label columns.

        Returns:
            Tabular pytorch dataset.
        """
        return cls(TabularDataset.from_csv(csv_path=csv_path, feature_cols=feature_cols, label_cols=label_cols))

    @classmethod
    def text_data_from_df(
        cls, df: pd.DataFrame, input_col: str, label_cols: Optional[Union[str, List[str]]], tokenizer, max_len: int,
    ) -> SimpleDataloader:
        """Classmethod to create a dataset as required by transformers for text classification tasks.

        Args:
            df: The dataframe containing sentences and labels.
            input_col: The name of column containing sentences.
            label_cols: name of label column, or a list containing names of label columns.
            tokenizer: The tokenizer to be used to tokenize the sentences.
            max_len: The max_length to be used by the tokenizer.

        Returns:
                pytorch dataset for text classification using huggingface.
        """
        return cls(
            TextClassificationDataset.from_df(
                df=df, input_col=input_col, label_cols=label_cols, tokenizer=tokenizer, max_len=max_len,
            )
        )

    def get_loader(self, batch_size: int = 32, shuffle: bool = True, **dl_params) -> DataLoader:
        """Method to get dataloader.

        Args:
            batch_size(int): The batch size to use
            shuffle(bool): Whether to shuffle the inputs.
            **dl_params(dict) : Keyword arguments related to dataloader

        Returns:
            A PyTorch dataloader with given arguments.
        """
        dl = DataLoader(self.ds, batch_size=batch_size, shuffle=shuffle, **dl_params)
        return dl



"""Implements AM-softmax."""
import torch
import torch.nn as nn
import torch.nn.functional as F


class AMSoftmax(nn.Module):
    """Implementation of Additive Margin Softmax for Face Verification.

    Additive Margin Softmax: <https://arxiv.org/abs/1801.05599>
    """

    def __init__(self, in_features, out_features, m=0.35, s=32):
        """Class Constructor.

        Args:
            in_features: Size of the input features
            out_features: The size of output features(usually number of num_classes)
            s: The norm for input features.
            m: margin
        """
        super(AMSoftmax, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.m = m
        self.s = s

        self.eps = 1e-7
        self.Weight = nn.Parameter(torch.FloatTensor(self.out_features, self.in_features))
        nn.init.xavier_uniform_(self.Weight)

    def forward(self, features: torch.Tensor, targets: torch.Tensor = None) -> torch.Tensor:
        """Forward Pass.

        Args:
            features: The input features of shape (BS x F) where BS is batch size and F is input feature dimension.
            targets: The targets with shape BS , where BS is batch size

        Returns:
            Logits with shape (BS x out_features)
        """
        cos_theta = F.linear(F.normalize(features), F.normalize(self.Weight))
        if targets is None:
            return cos_theta

        one_hot = torch.zeros_like(cos_theta)
        one_hot.scatter_(1, targets.view(-1, 1).long(), 1)

        logits = torch.where(one_hot.bool(), cos_theta - self.m, cos_theta)
        logits = torch.cos(logits)
        logits *= self.s

        return logits



"""Imports for notifiers."""
from torchflare.callbacks.notifiers.message_notifiers import DiscordNotifierCallback, SlackNotifierCallback

__all__ = ["DiscordNotifierCallback", "SlackNotifierCallback"]



"""Imports for experiment."""
from torchflare.experiments.criterion_utilities import get_criterion
from torchflare.experiments.experiment import Experiment
from torchflare.experiments.optim_utilities import get_optimizer
from torchflare.experiments.scheduler_utilities import LRScheduler, get_scheduler
from torchflare.experiments.simple_utils import to_device, to_numpy, wrap_metric_names
from torchflare.experiments.state import ExperimentState

__all__ = [
    "Experiment",
    "ExperimentState",
    "to_numpy",
    "to_device",
    "LRScheduler",
    "get_scheduler",
    "get_criterion",
    "get_optimizer",
    "wrap_metric_names",
]



"""Implements Accuracy-metric."""
import torch

from torchflare.metrics.meters import MetricMeter, _BaseMetric


class Accuracy(_BaseMetric, MetricMeter):
    """Computes Accuracy.

    Support binary,multilabel and multiclass cases
    """

    def __init__(self, num_classes: int, threshold: float = 0.5, multilabel: bool = False):
        """Constructor method for Accuracy Class.

        Args:
            num_classes: The number of num_classes.
            threshold: The threshold value to transform probability predictions to binary values(0,1)
            multilabel: Set it to True if your problem is  multilabel classification.
        """
        super(Accuracy, self).__init__(multilabel=multilabel)

        self.threshold = threshold
        self.num_classes = num_classes
        self._outputs = None
        self._targets = None

        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The class name
        """
        return self.__class__.__name__.lower()

    # noinspection PyTypeChecker
    def _compute(self, outputs: torch.Tensor, targets: torch.Tensor):

        self._check_type(outputs=outputs, targets=targets)
        if self.case_type == "multiclass":

            outputs = torch.argmax(outputs, dim=1)
            correct = torch.eq(outputs, targets).view(-1)

        else:

            outputs = torch.sigmoid(outputs)
            outputs = (outputs >= self.threshold).float()

            if self.case_type == "binary":
                outputs = outputs.view(-1)
                targets = targets.view(-1)
                correct = torch.eq(outputs, targets)
            else:
                last_dim = outputs.ndimension()
                outputs = torch.transpose(outputs, 1, last_dim - 1).reshape(-1, self.num_classes)
                targets = torch.transpose(targets, 1, last_dim - 1).reshape(-1, self.num_classes)
                correct = torch.all(targets == outputs.type_as(targets), dim=-1)
        # assert outputs.shape == targets.shape
        total = correct.shape[0]
        correct = torch.sum(correct)
        return correct, total

    def accumulate(self, outputs: torch.Tensor, targets: torch.Tensor):
        """Method to accumulate the outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : Ground truth targets
        """
        outputs, targets = self.detach_tensor(outputs), self.detach_tensor(targets)
        self._outputs.append(outputs)
        self._targets.append(targets)

    def compute(self) -> torch.Tensor:
        """Computes the Accuracy per epoch.

        Returns:
            The accuracy
        """
        outputs = torch.cat(self._outputs)
        targets = torch.cat(self._targets)
        correct, total = self._compute(outputs=outputs, targets=targets)
        return correct / total

    def reset(self):
        """Resets the accumulation lists."""
        self._outputs = []
        self._targets = []


__all__ = ["Accuracy"]



"""Implements Model History."""
import os
from abc import ABC
from typing import Dict

import pandas as pd

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class History(Callbacks, ABC):
    """Class to log metrics to console and save them to a CSV file."""

    def __init__(self):
        """Constructor class for History Class."""
        super(History, self).__init__(order=CallbackOrder.LOGGING)
        self.history = None

    def experiment_start(self):
        """Sets variables at experiment start."""
        self.history = {}

    def _update_history(self, logs):
        for key in logs:
            if key not in self.history:
                self.history[key] = []
                self.history[key].append(logs[key])
            else:
                self.history[key].append(logs[key])

    def _store_history(self):

        path = os.path.join(self.exp.save_dir, "history.csv")
        df = pd.DataFrame.from_dict(self.history)
        df.to_csv(path, index=False)

    def epoch_end(self, epoch: int, logs: Dict):
        """Method to update history object at the end of every epoch.

        Args:
            epoch: The current epoch.
            logs: Dictionary containing the metric and loss values.1
        """
        self._update_history(logs=dict(Epoch=epoch, **logs))

    def experiment_end(self):
        """Method to store history object at experiment enc."""
        self._store_history()



"""Implements dataset for Image segmentation tasks."""
from __future__ import annotations

import os
from typing import List, Optional, Tuple, Union

import albumentations as A
import pandas as pd
import torchvision
from torch.utils.data import Dataset

from torchflare.datasets.utils import apply_augmentations, make_masks, open_image


def get_rle(df: pd.DataFrame, image_col: str, mask_cols: List[str], name: str) -> List[str]:
    ids = df[df[image_col] == name].index.values
    rle = df.loc[ids, mask_cols].values.tolist()
    return rle


class SegmentationDataset(Dataset):
    """Class to create a dataset for image segmentation tasks."""

    def __init__(
        self,
        image_paths_list: List,
        mask_list=None,
        augmentations: [Union[A.Compose, torchvision.transforms.Compose]] = None,
        image_convert_mode: str = "RGB",
        mask_convert_mode: str = "L",
        **kwargs
    ):
        """Constructor method for segmentation Dataset.

        Args:
            image_paths_list:  A list containing paths to all images. Ex ['/train/img.png' , 'train/img_0.png' ,etc]
            mask_list: A list containing all the masks.
            augmentations: Augmentations to be applied to image , masks.
            image_convert_mode:  The mode in which images should be opened.
            mask_convert_mode: The mode in which masks should be opened.
            **kwargs: Extra named args.
        """
        self.image_list = image_paths_list
        self.mask_list = mask_list
        self.augmentations = augmentations
        self.kwargs = kwargs
        self.image_convert_mode = image_convert_mode
        self.mask_convert_mode = mask_convert_mode

    def __len__(self):
        """__len__ method.

        Returns:
            The length of dataloader.
        """
        return len(self.image_list)

    def __getitem__(self, item):
        """__getitem__ method.

        Args:
            item : The selected id.

        Returns:
            Tensors of images , masks if mask_list is provided else Tensors of images.
        """
        images = open_image(self.image_list[item], convert_mode=self.image_convert_mode)
        if self.mask_list is not None:
            if any(isinstance(ele, list) for ele in self.mask_list):
                mask = make_masks(self.mask_list[item], **self.kwargs)
            else:
                mask = open_image(self.mask_list[item], convert_mode=self.mask_convert_mode)
            images, mask = apply_augmentations(images, augs=self.augmentations, mask=mask)

            return images, mask

        else:

            images = apply_augmentations(images, augs=self.augmentations, mask=None)
            return images

    @staticmethod
    def _join_paths(path: str, file_names: List[str], extension: Optional[str] = None) -> List[str]:

        if extension is None:
            return [os.path.join(path, x) for x in file_names]

        return [os.path.join(path, x + extension) for x in file_names]

    @staticmethod
    def create_mask_list(df: pd.DataFrame, image_col: str, mask_cols: List[str]) -> List[List[str]]:
        """Create mask list.

        Args:
            df : The dataframe.
            image_col: The column containing image_names.
            mask_cols : The column/columns containing mask rle's.

        Returns:
            A list of masks
        """
        image_names = df[image_col].values.tolist()
        mask_list = []

        for name in image_names:
            mask_list.append(get_rle(df, image_col, mask_cols, name))

        return mask_list

    @classmethod
    def from_rle(
        cls,
        path: str,
        df: pd.DataFrame,
        image_col: str,
        mask_cols: List[str] = None,
        augmentations: Union[A.Compose, torchvision.transforms.Compose] = None,
        mask_size: Tuple[int, int] = None,
        num_classes: int = None,
        extension: str = None,
        image_convert_mode: str = "RGB",
    ) -> SegmentationDataset:
        """Classmethod to create pytorch dataset when you have rule length encodings for masks stored in a dataframe.

        Args:
            path: The path where images are saved.
            df: The dataframe containing the image name/ids, and the targets
            image_col: The name of the image column containing the image name/ids.
            augmentations: The batch_mixers to be used on images and the masks.
            mask_cols: The list of columns containing the rule length encoding.
            mask_size: The size of mask.
            num_classes: The number of num_classes.
            image_convert_mode: The mode to be passed to PIL.Image.convert.
            extension : The extension of image file.
                If your image_names do not have extension then set extension to '.jpg' or '.png' ,etc

        Returns:
            returns image_paths_list , mask_list , image_convert_mode , augmentations and extra kwargs.

        Note:
            If you want to create a dataset for testing set mask_cols = None, mask_size = None, num_classes = None.
            The created masks will be binary.
        """
        image_list = cls._join_paths(path=path, file_names=df[image_col].values.tolist(), extension=extension)

        mask_list = (
            cls.create_mask_list(df=df, image_col=image_col, mask_cols=mask_cols) if mask_cols is not None else None
        )

        return cls(
            image_list,
            mask_list,
            augmentations,
            image_convert_mode=image_convert_mode,
            shape=mask_size,
            num_classes=num_classes,
        )

    @classmethod
    def from_folders(
        cls,
        image_path: str,
        mask_path: str = None,
        augmentations: Union[A.Compose, torchvision.transforms.Compose] = None,
        image_convert_mode: str = "L",
        mask_convert_mode: str = "L",
    ) -> SegmentationDataset:
        """Classmethod to create pytorch dataset from folders.

        Args:
            image_path: The path where images are stored.
            mask_path: The path where masks are stored.
            augmentations: The batch_mixers to apply on images and masks.
            image_convert_mode: The mode to be passed to PIL.Image.convert for input images
            mask_convert_mode: The mode to be passed to PIL.Image.convert for masks.

        Returns:
            returns image_paths_list , mask_path , augmentations , image_convert_mode , mask_convert_mode.

        Note:
             If you want to create a dataset for testing just set mask_path = None.
        """
        image_files = cls._join_paths(image_path, os.listdir(image_path))
        mask_files = cls._join_paths(mask_path, os.listdir(mask_path)) if mask_path is not None else None

        return cls(
            image_paths_list=image_files,
            mask_list=mask_files,
            augmentations=augmentations,
            image_convert_mode=image_convert_mode,
            mask_convert_mode=mask_convert_mode,
        )


__all__ = ["SegmentationDataset"]



"""Implementation of Squeeze and Excitation BLocks."""

import torch
import torch.nn as nn
import torch.nn.functional as F


class CSE(nn.Module):
    """Implementation of Channel Wise Squeeze and Excitation Block.

    Paper : <https://arxiv.org/abs/1709.01507>

    Adapted from
    <https://www.kaggle.com/c/tgs-salt-identification-challenge/discussion/65939>
    and
    <https://www.kaggle.com/c/tgs-salt-identification-challenge/discussion/66178>
    """

    def __init__(self, in_channels: int, r: int = 16):
        """Constructor for CSE class.

        Args:
            in_channels : The number of input channels in the feature map.
            r : The reduction ration (Default : 16)
        """
        super(CSE, self).__init__()

        self.in_channels = in_channels
        self.r = r
        self.linear1 = nn.Linear(self.in_channels, self.in_channels // self.r)
        self.linear2 = nn.Linear(self.in_channels // r, self.in_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward Method.

        Args:
            x: The input tensor of shape (batch, channels, height, width)

        Returns:
            Tensor of same shape
        """
        x_inp = x

        x = x.view(*(x.shape[:-2]), -1).mean(-1)
        x = F.relu(self.linear1(x), inplace=True)
        x = self.linear2(x)
        x = x.unsqueeze(-1).unsqueeze(-1)
        x = torch.sigmoid(x)

        x = torch.mul(x_inp, x)

        return x


class SSE(nn.Module):
    """SSE : Channel Squeeze and Spatial Excitation block.

    Paper : <https://arxiv.org/abs/1803.02579>

    Adapted from
    <https://www.kaggle.com/c/tgs-salt-identification-challenge/discussion/66178>
    """

    def __init__(self, in_channels):
        """Constructor method for SSE class.

        Args:
            in_channels : The number of input channels in the feature map.
        """
        super(SSE, self).__init__()

        self.in_channels = in_channels
        # noinspection PyTypeChecker
        self.conv = nn.Conv2d(in_channels=self.in_channels, out_channels=1, kernel_size=1, stride=1)

    def forward(self, x) -> torch.Tensor:
        """Forward Method.

        Args:
            x: The input tensor of shape (batch, channels, height, width)

        Returns:
            Tensor of same shape
        """
        x_inp = x

        x = self.conv(x)
        x = torch.sigmoid(x)

        x = torch.mul(x_inp, x)

        return x


class SCSE(nn.Module):
    """Implementation of SCSE : Concurrent Spatial and Channel Squeeze and Channel Excitation block.

    Paper : <https://arxiv.org/abs/1803.02579>


    Adapted from
    <https://www.kaggle.com/c/tgs-salt-identification-challenge/discussion/66178>
    """

    def __init__(self, in_channels, r=16):
        """Constructor for SCSE class.

        Args:
            in_channels : The number of input channels in the feature map.
            r : The reduction ration (Default : 16)
        """
        super(SCSE, self).__init__()

        self.in_channels = in_channels
        self.r = r

        self.cse = CSE(in_channels=self.in_channels, r=self.r)
        self.sse = SSE(in_channels=self.in_channels)

    def forward(self, x) -> torch.Tensor:
        """Forward method.

        Args:
            x: The input tensor of shape (batch, channels, height, width)

        Returns:
            Tensor of same shape
        """
        cse = self.cse(x)
        sse = self.sse(x)

        op = torch.add(cse, sse)

        return op


__all__ = ["SSE", "SCSE", "CSE"]



"""Implements dataset for text classification task."""
from __future__ import annotations

from typing import List, Union

import pandas as pd
import torch
from torch.utils.data import Dataset


class TextClassificationDataset(Dataset):
    """Class to create a dataset for text classification as required by transformers."""

    def __init__(
        self, inputs: List[str], tokenizer, max_len: int, labels: [Union[str, List[str]]] = None,
    ):
        """Constructor Method.

        Args:
            inputs: A list of sentences.
            labels: A list of labels for classification.
            tokenizer: The huggingface tokenizer to be used.
            max_len : The max_length of ids.
        """
        self.inputs = inputs
        self.labels = labels
        self.tokenizer = tokenizer
        self.max_len = max_len

    def __len__(self):
        """__len__ method.

        Returns:
            The length of dataset.
        """
        return len(self.inputs)

    def __getitem__(self, item) -> Union:
        """__getitem__ method.

        Args:
            item : idx

        Returns:
            A dictionary containing input_ids , token_type_ids ,etc and the corresponding labels.
        """
        text = self.inputs[item]
        inps = self.tokenizer(
            text,
            add_special_tokens=True,
            max_length=self.max_len,
            padding="max_length",
            truncation=True,
            return_tensors="pt",
        )

        inps_dict = {k: inps[k].squeeze(0) for k in inps}  # Understand why extra dim is added by tokenizer

        if self.labels is not None:
            labels = self.labels[item]
            return inps_dict, torch.tensor(labels, dtype=torch.long)

        return inps_dict

    @classmethod
    def from_df(
        cls, df: pd.DataFrame, input_col: str, tokenizer, max_len: int, label_cols: Union[str, List[str]] = None,
    ) -> TextClassificationDataset:
        """Classmethod to create the dataset from dataframe.

        Args:
            df: The dataframe which has the data.
            input_col: The column containing the inputs.
            label_cols: The column which contains corresponding labels.
            tokenizer: The tokenizer to be used.(Use only tokenizer available in huggingface)
            max_len: The max_len to be used.

        Returns:
            A list of sentences and corresponding labels if label_cols is provided else return a list of sentences.
        """
        inputs = df.loc[:, input_col].values.tolist()

        labels = df.loc[:, label_cols].values.tolist() if label_cols is not None else None

        return cls(inputs=inputs, labels=labels, tokenizer=tokenizer, max_len=max_len)



"""Implements Tensorboard Logger."""
from abc import ABC

from torch.utils.tensorboard import SummaryWriter

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class TensorboardLogger(Callbacks, ABC):
    """Callback to use Tensorboard to log your metrics and losses."""

    def __init__(self, log_dir: str):
        """Constructor for TensorboardLogger class.

        Args:
            log_dir: The directory where you want to save your experiments.
        """
        super(TensorboardLogger, self).__init__(order=CallbackOrder.LOGGING)
        self.log_dir = log_dir
        self._experiment = SummaryWriter(log_dir=self.log_dir)

    def epoch_end(self, epoch: int, logs: dict):
        """Method to log metrics and values at the end of very epoch.

        Args:
            logs: A dictionary containing metrics and loss values.
            epoch: The current epoch
        """
        for key, value in logs.items():
            if not isinstance(value, str):
                self._experiment.add_scalar(tag=key, scalar_value=value, global_step=epoch)

    def experiment_end(self):
        """Method to end experiment after training is done."""
        self._experiment.close()



"""Implements Utility functions for datasets."""
from typing import Tuple

import albumentations as A
import matplotlib.pyplot as plt
import numpy as np
import PIL
import torch
import torchvision
import torchvision.transforms as transforms


def DecodeRLE(mask_rle: str, shape: Tuple):
    """A function to decode run length encoding.

    Args:
        mask_rle : rule length encoding
        shape : A tuple for the specified shape for mask
                Ex : shape = (128 , 128)

    Returns:
        Decoded encoding as numpy array.
    """
    # print(type(mask_rle))
    s = mask_rle.split()
    start, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
    start -= 1
    end = start + lengths
    img = np.zeros(shape[0] * shape[1], dtype=np.uint8)
    for lo_idx, hi_idx in zip(start, end):
        img[lo_idx:hi_idx] = 1

    return img.reshape(shape, order="F")


def EncodeRLE(mask):
    """Convert mask to EncodedPixels in run-length encoding.

    Source : https://www.kaggle.com/stainsby/fast-tested-rle-and-input-routines

    Args:
        mask : the mask array which you want to convert to Rule Length Encoding.

    Returns:
        RLE for the mask.
    """
    pixels = mask.T.flatten()
    # We need to allow for cases where there is a '1' at either end of the sequence.
    # We do this by padding with a zero at each end when needed.
    use_padding = False
    if pixels[0] or pixels[-1]:
        use_padding = True
        pixel_padded = np.zeros([len(pixels) + 2], dtype=pixels.dtype)
        pixel_padded[1:-1] = pixels
        pixels = pixel_padded
    rle = np.where(pixels[1:] != pixels[:-1])[0] + 2
    if use_padding:
        rle = rle - 1
    rle[1::2] = rle[1::2] - rle[:-1:2]
    return rle


def make_masks(rle, shape, num_classes):
    """Create masks based on RLE.

    Args:
        rle : rule length encoding
        shape : A tuple for the specified shape for mask
                Ex : shape = (128 , 128)
        num_classes : the number of masks if you have a multiclasss segmentation problem

    Returns:
        numpy array of masks.
    """
    masks = np.zeros((shape[0], shape[1], num_classes), dtype=np.float32)
    # print(len(rle))

    for idx, label in enumerate(rle):
        if isinstance(label, (list, tuple)):
            label = label[0]
        if isinstance(label, str):
            mask = DecodeRLE(label, shape=shape)
            masks[:, :, idx] = mask

    return masks


def handle_shape(x):
    """Handles shape for albumentations augs.

    Args:
        x : the input image.

    Returns:
        The image after handling the shape.
    """
    x = np.asarray(x)
    if x.ndim == 2:
        x = np.expand_dims(x, 2)
    x = np.transpose(x, (1, 0, 2))
    x = np.transpose(x, (2, 1, 0))

    return x


def open_image(x, convert_mode="RGB"):
    """Opens Image.

    Args:
        x : The path to image.
        convert_mode: The mode in which image has to be opened.

    Returns:
        PIL.Image in the required mode.
    """
    x = PIL.Image.open(x).convert(convert_mode)
    return x


def apply_albu_augs(image, augs, mask):
    """Applies albumentations augmentations.

    Args:
        image: The input image.
        augs: The albumentations augmentations.
        mask: The mask if problem type is image segmentation,

    Returns:
        image , mask if image segmentation else returns image.
    """
    image = np.array(image)
    if mask is not None:
        mask = np.array(mask)
        augmented_image = augs(image=image, mask=mask)
        mask = augmented_image.get("mask")

    else:
        augmented_image = augs(image=image)
        mask = None

    image = augmented_image.get("image")

    return image, mask


def apply_torchvision_augs(image, augs, mask):
    """Applies albumentations augmentations.

    Args:
        image: The input image.
        augs: The torchvision augmentations.
        mask: The mask if problem type is image segmentation,

    Returns:
        image , mask if image segmentation else returns image.
    """
    image = augs(image)
    if mask is not None:
        if isinstance(mask, np.ndarray):
            mask = mask.astype(np.uint8)
            mask = torchvision.transforms.ToPILImage()(mask)

        mask = augs(mask)
    else:
        mask = None
    return image, mask


def apply_augmentations(image, augs, mask=None):
    """Apply albumentations/torchvision augmentations.

    Args:
        image: The input image.
        augs: The augmentations.
        mask : The input mask(only applicable for image segmentation problems.)

    Returns:
        Image , mask if problem is image segmentation else returns image
    """
    if isinstance(augs, A.Compose):

        image, mask = apply_albu_augs(image=image, augs=augs, mask=mask)
        image = handle_shape(image)
        if mask is not None:
            mask = handle_shape(mask)

    elif isinstance(augs, torchvision.transforms.Compose):

        image, mask = apply_torchvision_augs(image=image, augs=augs, mask=mask)

    if mask is not None:

        return to_tensor(image), to_tensor(mask)

    else:

        return to_tensor(image)


def to_tensor(x):
    """Converts input to tensor.

    Args:
        x : input

    Returns:
        Torch.tensor
    """
    if not isinstance(x, torch.Tensor) and isinstance(x, PIL.Image.Image):

        return transforms.functional.to_tensor(x)
    elif not isinstance(x, torch.Tensor) and x is not None:
        return torch.tensor(x)
    else:
        return x


def show_batch(dl, **kwargs):
    """Method to visualize the batch for image data.

    Args:
        dl : The pytorch dataloader.

    Note:
        Only use for classification and segmentations tasks.
    """
    op = next(iter(dl))
    if isinstance(op, (list, tuple)):
        x, _ = op
    else:
        x = op
    grid_images = torchvision.utils.make_grid(x, **kwargs)
    grid_images = grid_images.numpy()
    plt.imshow(np.transpose(grid_images, (1, 2, 0)), interpolation="nearest")



"""Implements IOU Metric."""
import torch

from torchflare.metrics.meters import MetricMeter, calculate_segmentation_statistics


class IOU(MetricMeter):
    """Class which computes intersection over union."""

    def __init__(self, threshold: float = None, class_dim: int = 1):
        """Constructor method for IOU.

        Args:
            threshold: threshold for binarization of predictions
            class_dim: indicates class dimension (K)

        Note:
            Supports only binary cases
        """
        self.threshold = threshold
        self.class_dim = class_dim
        self.eps = 1e-20

        self._outputs = []
        self._targets = []
        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The class name
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs: torch.Tensor, targets: torch.Tensor):
        """Method to accumulate the outputs and targets.

        Args:
            outputs: [N, K, ...] tensor that for each of the N samples
                indicates the probability of the sample belonging to each of
                the K num_classes.
            targets:  binary [N, K, ...] tensor that encodes which of the K
                num_classes are associated with the N-th sample.
        """
        self._outputs.append(outputs)
        self._targets.append(targets)

    def compute(self) -> torch.Tensor:
        """Method to Compute IOU.

        Returns:
            The computed iou.
        """
        self._outputs = torch.cat(self._outputs)
        self._targets = torch.cat(self._targets)

        tp, fp, fn = calculate_segmentation_statistics(
            outputs=self._outputs, targets=self._targets, threshold=self.threshold, class_dim=self.class_dim,
        )

        union = tp + fp + fn

        score = (tp + self.eps * (union == 0).float()) / (tp + fp + fn + self.eps)

        return torch.mean(score)

    def reset(self):
        """Method to reset the accumulation lists."""
        self._outputs = []
        self._targets = []


__all__ = ["IOU"]



"""Implements container for loss and metric computation."""
from typing import Dict, List

from torchflare.experiments.simple_utils import wrap_metric_names
from torchflare.utils.average_meter import AverageMeter


class MetricAndLossContainer:
    """Class to run and compute Loss and Metrics."""

    def __init__(self, metrics: List = None):
        """Constructor class for MetricContainer.

        Args:
            metrics: The list of metrics
        """
        if metrics is not None:
            self.metrics = metrics
            self.metric_names = wrap_metric_names(metric_list=metrics)
        else:
            self.metrics = None

        self.exp = None
        self.loss_meter = AverageMeter()

    def set_experiment(self, exp):  # noqa
        self.exp = exp

    def reset(self):
        """Method to reset the state of metrics and loss meter."""
        self.loss_meter.reset()
        if self.metrics is not None and self.exp.compute_metric_flag:

            for metric in self.metrics:
                metric.reset()

    def compute(self, prefix: str) -> Dict:
        """Method to compute the metrics once accumulation of values is done.

        Args:
            prefix : The prefix to add to the keys.

        Returns:
            A dictionary containing corresponding metrics.
        """
        metric_dict = {prefix + "loss": self.loss_meter.avg}
        if self.metrics is not None and self.exp.compute_metric_flag:
            metric_vals = [metric.compute().item() for metric in self.metrics]
            metric_dict.update({prefix + key: value for key, value in zip(self.metric_names, metric_vals)})

        return metric_dict

    def accumulate(self, loss, n, op=None, y=None):
        """Method to accumulate output of every batch.

        Args:
            op : The output of the net.
            y : The corresponding targets/targets.
            loss : the loss per batch
            n : The number of batches in a single iteration(batch_size)
        """
        self.loss_meter.update(loss, n)
        if self.metrics is not None and self.exp.compute_metric_flag:
            for metric in self.metrics:
                metric.accumulate(op, y)


__all__ = ["MetricAndLossContainer"]



"""Definitions of experiment states and Callback order."""
from enum import Enum, IntEnum


class ExperimentStates(Enum):
    """Class Define various stages of Training."""

    EPOCH_START = "epoch_start"
    EPOCH_END = "epoch_end"
    EXP_START = "experiment_start"
    EXP_END = "experiment_end"
    BATCH_START = "batch_start"
    BATCH_END = "batch_end"
    STOP = "stop"


class CallbackOrder(IntEnum):
    """Callback orders."""

    INTERNAL = 0
    LOGGING = 1
    STOPPING = 2
    EXTERNAL = 3


__all__ = ["ExperimentStates", "CallbackOrder"]



"""Implements Experiment State."""
import os
import warnings
from typing import List

import matplotlib.pyplot as plt
import torch
from fastprogress.fastprogress import master_bar, progress_bar
from torch.utils.data import DataLoader

import torchflare.batch_mixers as special_augs
import torchflare.metrics.metric_utils as metric_utils
from torchflare.callbacks.callback import CallbackRunner, sort_callbacks
from torchflare.callbacks.load_checkpoint import LoadCheckpoint
from torchflare.callbacks.model_history import History
from torchflare.callbacks.timer import TimeCallback
from torchflare.experiments.criterion_utilities import get_criterion
from torchflare.experiments.optim_utilities import get_optimizer
from torchflare.experiments.scheduler_utilities import LRScheduler
from torchflare.experiments.simple_utils import wrap_metric_names
from torchflare.utils.seeder import seed_all


class ExperimentState:
    """Class to set user and internal variables along with some utility functions."""

    def __init__(
        self,
        num_epochs: int,
        save_dir: str,
        model_name: str,
        fp16: bool,
        device: str,
        compute_train_metrics: bool,
        using_batch_mixers: bool,
        seed: int = 42,
    ):
        """Init method to set up important variables for training and validation.

        Args:
            num_epochs : The number of epochs to save model.
            save_dir : The directory where to save the model, and the log files.
            model_name : The name of '.bin' file.
                Defaults to 'model.bin'
            fp16 : Set this to True if you want to use mixed precision training(Default : False)
            device : The device where you want train your model.
            compute_train_metrics: Whether to compute metrics on training data as well
            using_batch_mixers : Whether using special batch_mixers like cutmix or mixup.
            seed: The seed to ensure reproducibility.

        Note:
            If batch_mixers are used then set compute_train_metrics to False.
            Also, only validation metrics will be computed if special batch_mixers are used.
        """
        self.num_epochs = num_epochs
        self.save_dir = save_dir
        self.model_name = model_name
        self.fp16 = fp16
        self.device = device
        self.compute_train_metrics = compute_train_metrics
        self._using_batch_mixers = using_batch_mixers
        self.seed = seed
        self.train_key = "train_"
        self.val_key = "val_"
        self.epoch_key = "Epoch"

        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

        self.path = os.path.join(self.save_dir, self.model_name)
        self.scaler = torch.cuda.amp.GradScaler() if self.fp16 else None
        self.history = History()
        self.model = None
        self.resume_checkpoint = None
        self.main_metric = None
        self.stop_training = None
        self.experiment_state = None
        self._step_after = None
        self._callback_runner = None
        self.optimizer = None
        self.scheduler_stepper = None
        self.exp_logs = {}
        self._train_monitor = {}
        self._val_monitor = {}
        self.criterion = None
        self.compute_metric_flag = True
        self._metric_runner = None
        self.master_bar = None
        self.progress_bar = None
        self.header = None
        if self.compute_train_metrics is True and self._using_batch_mixers is True:
            warnings.warn(
                "Using special batch_mixers is set to True and compute train metrics is also set to true. "
                "Setting  compute train metrics to False."
            )

            self.compute_train_metrics = False

        self._compute_val_metrics = True
        self._train_dl = None
        self._valid_dl = None

        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

    def _write_stdout(self, stats):
        self.master_bar.write(
            [f"{val:.5f}" if isinstance(val, float) else str(val) for val in stats], table=True,
        )

    def _set_callbacks(self, callbacks: List):
        if callbacks is None:
            if self.resume_checkpoint:
                callbacks = [self.history, TimeCallback(), LoadCheckpoint()]
            else:
                callbacks = [self.history, TimeCallback()]
        else:
            if self.resume_checkpoint:
                callbacks = [self.history, TimeCallback(), LoadCheckpoint()] + callbacks
            else:
                callbacks = [self.history, TimeCallback()] + callbacks

        callbacks = sort_callbacks(callbacks)
        self._callback_runner = CallbackRunner(callbacks=callbacks)
        self._callback_runner.set_experiment(self)

    def _set_metrics(self, metrics):

        self._metric_runner = metric_utils.MetricAndLossContainer(metrics=metrics)
        self._metric_runner.set_experiment(self)
        self.header = self._create_metric_lists(metrics=metrics)

    def _set_params(self, optimizer_params):

        if "model_params" in optimizer_params:
            grad_params = optimizer_params.pop("model_params")
        else:
            grad_params = (param for param in self.model.parameters() if param.requires_grad)

        return grad_params

    def _set_optimizer(self, optimizer, optimizer_params):

        grad_params = self._set_params(optimizer_params=optimizer_params)
        if isinstance(optimizer, str):
            self.optimizer = get_optimizer(optimizer)(grad_params, **optimizer_params)
        else:
            self.optimizer = optimizer

    def _set_scheduler(self, scheduler, scheduler_params):
        if scheduler is not None:
            self.scheduler_stepper = LRScheduler(scheduler=scheduler, optimizer=self.optimizer, **scheduler_params)
            self.scheduler_stepper.set_experiment(self)

    def _set_criterion(self, criterion):

        self.criterion = get_criterion(criterion=criterion)
        if self._using_batch_mixers:
            self.criterion = special_augs.MixCriterion(criterion=self.criterion)

    def _create_metric_lists(self, metrics):
        train_l = [self.epoch_key, self.train_key + "loss"]
        val_l = [self.val_key + "loss"]
        if metrics is not None:
            metric_names = wrap_metric_names(metric_list=metrics)
            train_metrics = [(self.train_key + k) for k in metric_names] if self.compute_train_metrics else []
            val_metrics = [(self.val_key + k) for k in metric_names]
            train_l.extend(train_metrics)
            val_l.extend(val_metrics)

        train_l.extend(val_l)
        train_l.append("Time")
        return train_l

    @property
    def set_state(self):
        """Returns the current state of experiment.

        Returns:
            The current experiment state.
        """
        return self.experiment_state

    @set_state.setter
    def set_state(self, state: str):

        self.experiment_state = state
        # Run callbacks on state change
        epoch = self.exp_logs.pop(self.epoch_key) if self.epoch_key in self.exp_logs.keys() else None
        self._callback_runner(current_state=self.experiment_state, epoch=epoch, logs=self.exp_logs)

    def _model_to_device(self):
        """Function to move model to device."""
        if next(self.model.parameters()).is_cuda is False:
            self.model.to(self.device)

    def _reset_model_logs(self):
        if bool(self.exp_logs):
            self.exp_logs = {}

    def initialize(self, train_dl: DataLoader, valid_dl: DataLoader):
        """Function to move model to device and seed everything.

        Args:
            train_dl : The training dataloader.
            valid_dl : The validation dataloader.
        """
        seed_all(self.seed)
        self._model_to_device()
        self._reset_model_logs()
        self.master_bar = master_bar(range(self.num_epochs))
        self._train_dl = train_dl
        self._valid_dl = valid_dl
        self._write_stdout(stats=self.header)

    def _create_bar(self, iterator):
        """Function to create progress bar.

        Args:
            iterator: The iterator to be used for progress bar.
        """
        self.progress_bar = progress_bar(iterator, parent=self.master_bar, leave=False)
        self.progress_bar.update(0)

    def _update_pbar(self, prefix, val):
        self.progress_bar.comment = f"{prefix + 'loss'} : {str(val)}"

    def _before_step(self, iterator, prefix):
        # before every train/val step create progress bar.
        self._create_bar(iterator=iterator)
        self.compute_metric_flag = self.compute_train_metrics if "train_" in prefix else self._compute_val_metrics

    def cleanup(self):
        """Function to reset the states of monitors and model."""
        self._train_monitor, self._val_monitor, self.exp_logs = {}, {}, {}
        self.experiment_state = None
        self.master_bar = None

    def _run_event(self, event: str, **kwargs):
        _ = getattr(self, event)(**kwargs)

    def plot_history(self, key: str, save_fig: bool = False, plot_fig: bool = True):
        """Method to plot model history.

        Args:
            key : A key value in lower case. Ex accuracy or loss
            save_fig: Set to True if you want to save_fig.
            plot_fig: Whether to plot the figure.
        """
        plt.figure(figsize=(10, 10))
        for k, v in self.history.history.items():
            if key in k:
                plt.plot(v, label=k)
        plt.title(f"{key}/{self.epoch_key}")
        plt.ylabel(f"{key}")
        plt.xlabel(self.epoch_key)
        plt.legend(loc="upper left")

        if save_fig is not None:
            save_path = os.path.join(self.save_dir, f"{key}-vs-{self.epoch_key}.jpg")
            plt.savefig(save_path, dpi=150)
        if plot_fig:
            plt.show()


__all__ = ["ExperimentState"]



"""Implements logger for weights and biases."""
from abc import ABC
from typing import Dict, List, Optional

import wandb

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class WandbLogger(Callbacks, ABC):
    """Callback to log your metrics and loss values to  wandb platform.

    For more information about wandb take a look at [Weights and Biases](https://wandb.ai/)
    """

    def __init__(
        self,
        project: str,
        entity: str,
        name: str = None,
        config: Dict = None,
        tags: List[str] = None,
        notes: Optional[str] = None,
        directory: str = None,
    ):
        """Constructor of WandbLogger.

        Args:
            project: The name of the project where you're sending the new run
            entity:  An entity is a username or team name where you're sending runs.
            name: A short display name for this run
            config: The hyperparameters for your model and experiment as a dictionary
            tags:  List of strings.
            directory: where to save wandb local run directory.
                If set to None it will use experiments save_dir argument.
            notes: A longer description of the run, like a -m commit message in git

        Note:
            set os.environ['WANDB_SILENT'] = True to silence wandb log statements.
            If this is set all logs will be written to WANDB_DIR/debug.log

        """
        super(WandbLogger, self).__init__(order=CallbackOrder.LOGGING)
        self.experiment = wandb.init(
            entity=entity, project=project, name=name, config=config, tags=tags, notes=notes, dir=directory
        )

    def epoch_end(self, epoch, logs):
        """Method to log metrics and values at the end of very epoch.

        Args:
            logs: A dictionary containing metrics and loss values.
            epoch: The current epoch
        """
        _ = logs.pop("Time")
        self.experiment.log(logs)

    def experiment_end(self):
        """Method to end experiment after training is done."""
        self.experiment.finish()



"""Implements Image datasets for classification."""
from __future__ import annotations

import os
from typing import Dict, List, Optional, Tuple, Union

import albumentations as A
import pandas as pd
import torchvision
from torch.utils.data import Dataset

from torchflare.datasets.utils import apply_augmentations, open_image, to_tensor


def get_files(directory: str):
    directory = os.path.expanduser(directory)
    return [os.path.join(directory, fname) for fname in os.listdir(directory)]


def get_files_and_labels(directory: str, class_to_idx: Dict[str, int]) -> Tuple[List, List]:
    image_list = []
    labels = []

    directory = os.path.expanduser(directory)

    for target_class in sorted(class_to_idx.keys()):
        class_idx = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)

        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                image_list.append(path)
                labels.append(class_idx)

    return image_list, labels


class ImageDataset(Dataset):
    """Class to create the dataset for Image Classification."""

    def __init__(
        self,
        image_paths_list: List,
        label_list: Optional[List] = None,
        augmentations: Optional[Union[A.Compose, torchvision.transforms.Compose]] = None,
        convert_mode: str = "RGB",
    ):
        """Constructor method for Image Dataset class.

        Args:
            image_paths_list: A list containing paths to all images. Ex ['/train/img.png' , 'train/img_0.png' ,etc]
            label_list: A list containing all the labels.
            augmentations : The augmentations to apply to the images.
            convert_mode: The mode in which the image is opened.
        """
        self.label_list = label_list
        self.image_list = image_paths_list
        self.augmentations = augmentations
        self.convert_mode = convert_mode

    @staticmethod
    def _get_labels_from_folders(path: str):
        classes = [d.name for d in os.scandir(path) if d.is_dir()]
        classes.sort()
        return {class_name: i for i, class_name in enumerate(classes)}

    @staticmethod
    def _get_labels_from_df(df: pd.DataFrame, label_cols: Union[str, List[str]]):

        return df.loc[:, label_cols].values.tolist() if label_cols is not None else None

    @staticmethod
    def _join_paths(path: str, file_names: List[str], extension: Optional[str] = None) -> List[str]:

        if extension is None:
            return [os.path.join(path, x) for x in file_names]

        return [os.path.join(path, x + extension) for x in file_names]

    @classmethod
    def from_df(
        cls,
        path: str,
        df: pd.DataFrame,
        image_col: str,
        label_cols: Union[str, List[str]] = None,
        augmentations: Union[A.Compose, torchvision.transforms.Compose] = None,
        extension: str = None,
        convert_mode: str = "RGB",
    ) -> ImageDataset:
        """Classmethod to create pytorch dataset from the given dataframe.

        Args:
            path: The path where images are saved.
            df: The dataframe containing the image name/ids, and the targets
            image_col: The name of the image column containing the image name/ids along with image extension.
                i.e. the images should have names like img_215.jpg or img_name.png ,etc
            augmentations: The batch_mixers to be used on images.
            label_cols: Column name or list of column names containing targets.
            extension : The image file extension.
            convert_mode: The mode to be passed to PIL.Image.convert.

        Returns:
            return image_paths_list , labels_list , augmentations and convert_mode.

        Note:
            For inference do not pass in the label_cols, keep it None.

            Augmentations : They must be Compose objects from albumentations or torchvision.
                When using albumentations do not use ToTensorV2().

            extension : If you specify extension be it jpg,png,etc. Please include '.' in extension
                i.e. '.jpg' or '.png'.
        """
        img_list = cls._join_paths(path=path, file_names=df.loc[:, image_col].values, extension=extension)
        label_list = cls._get_labels_from_df(df=df, label_cols=label_cols)
        return cls(
            image_paths_list=img_list, label_list=label_list, augmentations=augmentations, convert_mode=convert_mode,
        )

    @classmethod
    def from_folders(
        cls,
        path: str,
        augmentations: Union[A.Compose, torchvision.transforms.Compose] = None,
        convert_mode: str = "RGB",
    ) -> ImageDataset:
        """Classmethod to create pytorch dataset from folders.

        Args:
            path: The path where images are stored.
            augmentations:The batch_mixers to be used on images.
            convert_mode: The mode to be passed to PIL.Image.convert.

        Returns:
            return image_paths_list , labels_list , augmentations and convert_mode.

        Note:
            Augmentations must be Compose objects from albumentations or torchvision.

            The training directory structure should be as follows:
                train/class_1/xxx.jpg
                .
                .
                train/class_n/xxz.jpg

            The test directory structure should be as follows:
                test_dir/xxx.jpg
                test_dir/xyz.jpg
                test_dir/ppp.jpg
        """
        class_to_idx = cls._get_labels_from_folders(path)
        if class_to_idx:
            image_list, label_list = get_files_and_labels(path, class_to_idx)
        else:
            image_list, label_list = get_files(path), None

        return cls(
            image_paths_list=image_list, label_list=label_list, augmentations=augmentations, convert_mode=convert_mode,
        )

    def __len__(self):
        """__len__ method.

        Returns:
            length of dataloader.
        """
        return len(self.image_list)

    def __getitem__(self, item: int):
        """__getitem__ method.

        Args:
            item: The id

        Returns:
            Tensors of Image , labels if labels are present else Tensors of Images.
        """
        images = open_image(self.image_list[item], convert_mode=self.convert_mode)
        images = apply_augmentations(images, augs=self.augmentations, mask=None)
        if self.label_list is None:
            return images

        labels = self.label_list[item]
        return images, to_tensor(labels)


__all__ = ["ImageDataset"]



"""Imports for criterion."""
from torchflare.criterion.cross_entropy import BCEFlat, BCEWithLogitsFlat, LabelSmoothingCrossEntropy, SymmetricCE
from torchflare.criterion.dice_loss import DiceLoss
from torchflare.criterion.focal_loss import BCEFocalLoss, FocalCosineLoss, FocalLoss
from torchflare.criterion.iou_loss import IOULoss
from torchflare.criterion.triplet_loss import TripletLoss

__all__ = [
    "DiceLoss",
    "LabelSmoothingCrossEntropy",
    "SymmetricCE",
    "FocalLoss",
    "FocalCosineLoss",
    "IOULoss",
    "TripletLoss",
    "BCEFocalLoss",
    "BCEWithLogitsFlat",
    "BCEFlat",
]



"""Imports for datasets."""
from torchflare.datasets.classification import ImageDataset
from torchflare.datasets.dataloaders import SimpleDataloader
from torchflare.datasets.segmentation import SegmentationDataset
from torchflare.datasets.tabular import TabularDataset
from torchflare.datasets.text_dataset import TextClassificationDataset
from torchflare.datasets.utils import show_batch

__all__ = [
    "ImageDataset",
    "TabularDataset",
    "SegmentationDataset",
    "SimpleDataloader",
    "TextClassificationDataset",
    "show_batch",
]



"""Imports for loggers."""
from torchflare.callbacks.logging.comet_logger import CometLogger
from torchflare.callbacks.logging.neptune_logger import NeptuneLogger
from torchflare.callbacks.logging.tensorboard_logger import TensorboardLogger
from torchflare.callbacks.logging.wandb_logger import WandbLogger

__all__ = ["CometLogger", "NeptuneLogger", "TensorboardLogger", "WandbLogger"]



"""Implements DiceLoss."""
import torch
import torch.nn as nn

from torchflare.metrics.dice_meter import DiceScore


class DiceLoss(nn.Module):
    """Implementation of Dice Loss."""

    def __init__(self, class_dim=1):
        """Constructor method for Dice Loss.

        Args:
            class_dim: The dimension indication class.
        """
        super(DiceLoss, self).__init__()
        self.dice = DiceScore(threshold=None, class_dim=class_dim)

    def forward(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Forward method.

        Args:
            outputs: outputs from the net after applying activations.
            targets: The targets.

        Returns:
            The computed loss value.
        """
        self.dice.reset()
        self.dice.accumulate(outputs=outputs, targets=targets)
        return 1 - self.dice.compute()


__all__ = ["DiceLoss"]



"""Implements LiArcFace."""
import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class LiArcFace(nn.Module):
    """Implementation of Li-ArcFace.

    AirFace: [Lightweight and Efficient Model for Face Recognition](https://arxiv.org/abs/1907.12256)
    """

    def __init__(self, in_features, out_features, s=64, m=0.45):
        """Constructor class of LiArcFace.

        Args:
            in_features: Size of the input features
            out_features: The size of output features(usually number of num_classes)
            s: The norm for input features.
            m: margin
        """
        super(LiArcFace, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.s = s
        self.m = m
        self.eps = 1e-7
        self.Weight = nn.Parameter(torch.FloatTensor(self.out_features, self.in_features))
        nn.init.xavier_uniform_(self.Weight)

    def forward(self, features: torch.Tensor, targets: torch.Tensor = None) -> torch.Tensor:
        """Forward Pass.

        Args:
            features: The input features of shape (BS x F) where BS is batch size and F is input feature dimension.
            targets: The targets with shape BS , where BS is batch size

        Returns:
            Logits with shape (BS x out_features)
        """
        cos_theta = F.linear(F.normalize(features), F.normalize(self.Weight))
        if targets is None:
            return cos_theta

        cos_theta.clamp(-1 + self.eps, 1 - self.eps)

        theta = torch.acos(cos_theta)

        one_hot = torch.zeros_like(cos_theta)
        one_hot.scatter_(1, targets.data.view(-1, 1), 1)
        target = (math.pi - 2 * (theta + self.m)) / math.pi
        other = (math.pi - 2 * theta) / math.pi

        output = (one_hot * target) + ((1.0 - one_hot) * other)

        output = output * self.s
        return output


__all__ = ["LiArcFace"]



"""Implements notifiers for slack and discord."""
import json
from abc import ABC
from typing import Dict

import requests

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


def prepare_data(logs: dict):
    """Function to prepare the data according to the type of message.

    Args:
        logs: Dictionary containing the metrics and loss values.

    Returns:
        string in the same format as logs.
    """
    val = [f"{key} : {value}" for key, value in logs.items()]
    text = "\n".join(val)
    return text


class SlackNotifierCallback(Callbacks, ABC):
    """Class to Dispatch Training progress to your Slack channel."""

    def __init__(self, webhook_url: str):
        """Constructor method for SlackNotifierCallback.

        Args:
            webhook_url : Slack webhook url
        """
        super(SlackNotifierCallback, self).__init__(order=CallbackOrder.EXTERNAL)
        self.webhook_url = webhook_url

    def epoch_end(self, epoch: int, logs: Dict):
        """This function will dispatch messages to your Slack channel.

        Args:
            epoch : The current epoch
            logs : a str, list of strings, dictionary

        Raises:
            ValueError: If connection to slack channel could not be established.
        """
        logs = {"Epoch": epoch, **logs}
        data = {"text": prepare_data(logs)}

        response = requests.post(self.webhook_url, json.dumps(data), headers={"Content-Type": "application/json"})

        if response.status_code != 200:
            raise ValueError(
                "Request to slack returned an error {}, the response is:\n{}".format(
                    response.status_code, response.text
                )
            )


class DiscordNotifierCallback(Callbacks, ABC):
    """Class to Dispatch Training progress to your Discord Sever."""

    def __init__(self, exp_name: str, webhook_url: str):
        """Constructor method for DiscordNotifierCallback.

        Args:
            exp_name : The name of your experiment bot. (Can be anything)
            webhook_url : The webhook url of your discord server/channel.
        """
        super(DiscordNotifierCallback, self).__init__(order=CallbackOrder.EXTERNAL)
        self.exp_name = exp_name
        self.webhook_url = webhook_url

    def epoch_end(self, epoch: int, logs: Dict):
        """This function will dispatch messages to your discord server/channel.

        Args:
            epoch : The current epoch
            logs : a str, list of strings, dictionary
        """
        logs = {"Epoch": epoch, **logs}
        data = {
            "username": self.exp_name,
            "embeds": [{"description": prepare_data(logs)}],
        }
        response = requests.post(self.webhook_url, json.dumps(data), headers={"Content-Type": "application/json"})

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            print(err)



"""Implements get criterion method."""
import torch.nn.functional as F


def get_criterion(criterion):
    """Method to get criterion from nn.functional.

    Args:
        criterion: The criterion either str or callable.

    Returns:
        The Required criterion.

    Raises:
        ValueError: If input is string and criterion is not found in nn.functional value error is raised.
    """
    dir_f = dir(F)
    loss_fns = [d.lower() for d in dir_f]

    if isinstance(criterion, str):
        try:
            idx = loss_fns.index(criterion.lower())
            crit = getattr(F, dir(F)[idx])
            return crit
        except ValueError:
            raise ValueError("Invalid loss string input - must match pytorch function in torch.nn.functional")

    elif callable(criterion):

        return criterion



"""Implements FBeta and F1-score."""
import torch

from torchflare.metrics.meters import MetricMeter, _BaseInputHandler


class FBeta(_BaseInputHandler, MetricMeter):
    """Computes Fbeta Score.

    Supports binary,multiclass and multilabel cases.
    """

    def __init__(
        self, beta: float, num_classes: int, threshold: float = 0.5, average: str = "macro", multilabel: bool = False,
    ):
        """Constructor method for Fbeta score.

        Args:
            num_classes : The number of num_classes(For binary case , use out_features : 1)
            threshold: The value of threshold for masking. Input is raw logits.
            average : One of "micro" or "macro"
            beta : weight of precision in harmonic mean.
            multilabel: Whether problem is multilabel or not.

        Note:
            In case of binary classification, set num_classes = 1
        """
        super(FBeta, self).__init__(
            num_classes=num_classes, multilabel=multilabel, threshold=threshold, average=average,
        )

        self.beta = beta
        self.eps = 1e-20

        self._outputs = None
        self._targets = None

        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The class name
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs: torch.Tensor, targets: torch.Tensor):
        """Method to accumulate the outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : Ground truth targets
        """
        outputs, targets = self.detach_tensor(outputs), self.detach_tensor(targets)

        self._outputs.append(outputs)
        self._targets.append(targets)

    def reset(self):
        """Resets the accumulation lists."""
        self._outputs = []
        self._targets = []

    def compute(self) -> torch.Tensor:
        """Computes the FBeta Score.

        Returns:
            The computed Fbeta score.
        """
        outputs = torch.cat(self._outputs)
        targets = torch.cat(self._targets)

        tp, fp, tn, fn = self.compute_stats(outputs=outputs, targets=targets)

        precision = tp / (tp + fp + self.eps)
        recall = tp / (tp + fn + self.eps)

        numerator = (1 + self.beta ** 2) * precision * recall
        denominator = self.beta ** 2 * precision + recall

        fbeta = self.reduce(numerator=numerator, denominator=denominator)
        return fbeta


class F1Score(_BaseInputHandler, MetricMeter):
    """Computes F1 Score.

    Supports binary,multiclass and multilabel cases.
    """

    def __init__(
        self, num_classes: int, threshold: float = 0.5, multilabel: bool = False, average: str = "macro",
    ):
        """Constructor method for F1-score.

        Args:
            num_classes : The number of num_classes(For binary case , use out_features : 1)
            threshold: The value of threshold for masking. Input is raw logits.
            average : One of "micro" or "macro".
            multilabel: Whether the problem is multilabel or not.
        """
        super(F1Score, self).__init__(
            num_classes=num_classes, multilabel=multilabel, threshold=threshold, average=average,
        )

        self.eps = 1e-20
        self._outputs = None
        self._targets = None

        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The class name
        """
        return self.__class__.__name__.lower()

    def compute(self) -> torch.Tensor:
        """Method to compute FBeta Score.

        Returns:
            The computed F1-score
        """
        outputs = torch.cat(self._outputs)
        targets = torch.cat(self._targets)

        tp, fp, tn, fn = self.compute_stats(outputs=outputs, targets=targets)

        precision = tp / (tp + fp + self.eps)
        recall = tp / (tp + fn + self.eps)

        numerator = 2 * precision * recall
        denominator = precision + recall

        f1 = self.reduce(numerator=numerator, denominator=denominator)
        return f1

    def accumulate(self, outputs: torch.Tensor, targets: torch.Tensor):
        """Method to accumulate the outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : Ground truth targets
        """
        outputs, targets = self.detach_tensor(outputs), self.detach_tensor(targets)

        self._outputs.append(outputs)
        self._targets.append(targets)

    def reset(self):
        """Resets the accumulation lists."""
        self._outputs = []
        self._targets = []


__all__ = ["FBeta", "F1Score"]



"""Imports for callbacks."""
from torchflare.callbacks.callback import CallbackRunner, Callbacks, sort_callbacks
from torchflare.callbacks.early_stopping import EarlyStopping
from torchflare.callbacks.load_checkpoint import LoadCheckpoint
from torchflare.callbacks.logging.comet_logger import CometLogger  # noqa
from torchflare.callbacks.logging.neptune_logger import NeptuneLogger  # noqa
from torchflare.callbacks.logging.tensorboard_logger import TensorboardLogger  # noqa
from torchflare.callbacks.logging.wandb_logger import WandbLogger  # noqa
from torchflare.callbacks.model_checkpoint import ModelCheckpoint
from torchflare.callbacks.model_history import History
from torchflare.callbacks.notifiers.message_notifiers import DiscordNotifierCallback, SlackNotifierCallback
from torchflare.callbacks.states import CallbackOrder, ExperimentStates
from torchflare.callbacks.timer import TimeCallback

__all__ = [
    "Callbacks",
    "CallbackOrder",
    "CallbackRunner",
    "sort_callbacks",
    "ExperimentStates",
    "EarlyStopping",
    "SlackNotifierCallback",
    "DiscordNotifierCallback",
    "CometLogger",
    "NeptuneLogger",
    "TimeCallback",
    "TensorboardLogger",
    "WandbLogger",
    "ModelCheckpoint",
    "History",
    "LoadCheckpoint",
]



"""Implements triplet loss."""
import torch
import torch.nn as nn
import torch.nn.functional as F

from torchflare.criterion.utils import cosine_dist, euclidean_dist


def softmax_weights(dist, mask):

    max_v = torch.max(dist * mask, dim=1, keepdim=True)[0]
    difference = dist - max_v
    z = torch.sum(torch.exp(difference) * mask, dim=1, keepdim=True) + 1e-6  # avoid division by zero
    weights = torch.exp(difference) * mask / z
    return weights


# Adapted From : https://github.com/earhian/Humpback-Whale-Identification-1st-/blob/master/models/triplet_loss.py
def hard_example_mining(distance_matrix, pos_idxs, neg_idxs):
    """For each anchor, find the hardest positive and negative sample.

    Args:
        distance_matrix: pair wise distance between samples, shape [N, M]
        pos_idxs: positive index with shape [N, M]
        neg_idxs: negative index with shape [N, M]

    Returns:
        dist_ap: pytorch Variable, distance(anchor, positive); shape [N]
        dist_an: pytorch Variable, distance(anchor, negative); shape [N]
        p_inds: pytorch LongTensor, with shape [N];
            indices of selected hard positive samples; 0 <= p_inds[i] <= N - 1
        n_inds: pytorch LongTensor, with shape [N];
            indices of selected hard negative samples; 0 <= n_inds[i] <= N - 1

    Note:
        Only consider the case in which all targets have same num of samples,
        thus we can cope with all anchors in parallel.
    """
    assert len(distance_matrix.size()) == 2  # noqa: S101

    # `dist_ap` means distance(anchor, positive)
    # both `dist_ap` and `relative_p_inds` with shape [N]
    dist_ap, _ = torch.max(distance_matrix * pos_idxs, dim=1)
    # `dist_an` means distance(anchor, negative)
    # both `dist_an` and `relative_n_inds` with shape [N]
    dist_an, _ = torch.min(distance_matrix * neg_idxs + pos_idxs * 99999999.0, dim=1)

    return dist_ap, dist_an


def weighted_example_mining(distance_matrix, pos_idxs, neg_idxs):
    """For each anchor, find the weighted positive and negative sample.

    Args:
        distance_matrix: pytorch Variable, pair wise distance between samples, shape [N, N]
        pos_idxs:positive index with shape [N, M]
        neg_idxs: negative index with shape [N, M]

    Returns:
        dist_ap: pytorch Variable, distance(anchor, positive); shape [N]
        dist_an: pytorch Variable, distance(anchor, negative); shape [N]
    """
    assert len(distance_matrix.size()) == 2  # noqa: S101

    dist_ap = distance_matrix * pos_idxs
    dist_an = distance_matrix * neg_idxs

    weights_ap = softmax_weights(dist_ap, pos_idxs)
    weights_an = softmax_weights(-dist_an, neg_idxs)

    dist_ap = torch.sum(dist_ap * weights_ap, dim=1)
    dist_an = torch.sum(dist_an * weights_an, dim=1)

    return dist_ap, dist_an


class TripletLoss(nn.Module):
    """Computes Triplet loss."""

    def __init__(
        self, normalize_features: bool = True, margin: float = None, hard_mining: bool = True,
    ):
        """Constructor method for TripletLoss.

        Args:
            normalize_features: Whether to normalize the features. Default = True
            margin: The value for margin. Default = None.
            hard_mining: Whether to use hard sample mining. Default = True.
        """
        super(TripletLoss, self).__init__()

        self.normalize_features = normalize_features
        self.margin = margin
        self.hard_mining = hard_mining

    def forward(self, embedding: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Forward Method.

        Args:
            embedding: The output of the network.
            targets: The targets.

        Returns:
            The computed Triplet Loss.
        """
        distance_matrix = (
            cosine_dist(embedding, embedding) if self.normalize_features else euclidean_dist(embedding, embedding)
        )

        n = distance_matrix.size(0)
        pos_idxs = targets.view(n, 1).expand(n, n).eq(targets.view(n, 1).expand(n, n).t()).float()
        neg_idxs = targets.view(n, 1).expand(n, n).ne(targets.view(n, 1).expand(n, n).t()).float()

        if self.hard_mining:

            dist_ap, dist_an = hard_example_mining(
                distance_matrix=distance_matrix, pos_idxs=pos_idxs, neg_idxs=neg_idxs
            )

        else:

            dist_ap, dist_an = weighted_example_mining(
                distance_matrix=distance_matrix, pos_idxs=pos_idxs, neg_idxs=neg_idxs
            )

        y = dist_an.new().resize_as_(dist_an).fill_(1)

        if self.margin is not None and self.margin > 0:
            loss = F.margin_ranking_loss(dist_an, dist_ap, y, margin=self.margin)
        else:
            loss = F.soft_margin_loss(dist_an - dist_ap, y)
            # fmt: off
            if loss == float("Inf"):
                loss = F.margin_ranking_loss(dist_an, dist_ap, y, margin=0.3)
            # fmt: on

        return loss


__all__ = ["TripletLoss"]



"""Implementation of Early stopping."""
from abc import ABC
from typing import Dict

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class EarlyStopping(Callbacks, ABC):
    """Implementation of Early Stopping Callback."""

    def __init__(
        self, monitor: str = "val_loss", patience: int = 5, mode: str = "min", min_delta: float = 1e-7,
    ):
        """Constructor for EarlyStopping class.

        Args:
            monitor: The quantity to be monitored. (Default : val_loss)
                    If you want to monitor other metric just pass in the name of the metric.
            patience: Number of epochs with no improvement after which training will be stopped.
            mode: One of {"min", "max"}. In min mode, training will stop when the quantity monitored
                has stopped decreasing.In "max" mode it will stop when the quantity monitored has stopped increasing.
            min_delta: Minimum change in the monitored quantity to qualify as an improvement.

        Note:

            EarlyStopping will only use the values of metrics/loss obtained on validation set.
        """
        super(EarlyStopping, self).__init__(order=CallbackOrder.STOPPING)

        if "val_" not in monitor:
            self.monitor = "val_" + monitor
        else:
            self.monitor = monitor
        self.patience = patience
        self.mode = mode
        self.min_delta = min_delta
        self.stopping_counter = 0
        self.best_score = None
        self.improvement = None
        if self.mode == "min":

            self.improvement = lambda score, best_score: score <= (best_score - self.min_delta)

        else:

            self.improvement = lambda score, best_score: score >= (best_score + self.min_delta)

        self.stopping_counter = 0

    def epoch_end(self, epoch: int, logs: Dict):
        """Function which will determine when to stop the training depending on the score.

        Args:
            logs: A dictionary containing metrics and loss values.
            epoch : The current epoch
        """
        epoch_score = logs.get(self.monitor)
        if self.best_score is None or self.improvement(score=epoch_score, best_score=self.best_score):

            self.best_score = epoch_score

        else:

            self.stopping_counter += 1

        if self.stopping_counter >= self.patience:
            print("Early Stopping !")
            self.exp.stop_training = True



"""Imports for modules."""
from torchflare.modules.airface import LiArcFace
from torchflare.modules.am_softmax import AMSoftmax
from torchflare.modules.arcface import ArcFace
from torchflare.modules.cosface import CosFace
from torchflare.modules.se_modules import CSE, SCSE, SSE

__all__ = ["LiArcFace", "AMSoftmax", "ArcFace", "CosFace", "SCSE", "CSE", "SSE"]



"""Implements Average Meter."""


class AverageMeter:
    """Computes and stores the average and current value."""

    def __init__(self):
        """Constructor class for Average Meter."""
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        """Method to reset all the internal variables."""
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """Method to update values.

        Args:
            val: The current values.
            n: The count.
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



"""Implements variants for Cross Entropy loss."""
import torch
import torch.nn as nn
import torch.nn.functional as F


def BCEWithLogitsFlat(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Same as F.binary_cross_entropy_with_logits but flattens the input and target.

    Args:
        x : logits
        y: The corresponding targets.

    Returns:
        The computed Loss
    """
    y = y.view(x.shape).type_as(x)
    return torch.nn.functional.binary_cross_entropy_with_logits(x, y)


def BCEFlat(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Same as F.binary_cross_entropy but flattens the input and target.

    Args:
        x : logits
        y: The corresponding targets.

    Returns:
        The computed Loss
    """
    x = torch.sigmoid(x)
    y = y.view(x.shape).type_as(x)
    return torch.nn.functional.binary_cross_entropy(x, y)


class LabelSmoothingCrossEntropy(nn.Module):
    """NLL loss with targets smoothing."""

    def __init__(self, smoothing: float = 0.1):
        """Constructor method for LabelSmoothingCrossEntropy.

        Args:
            smoothing : targets smoothing factor

        Raises:
            ValueError: value error is raised if smoothing  > 1.0.
        """
        super(LabelSmoothingCrossEntropy, self).__init__()
        if smoothing > 1.0:
            raise ValueError("Smoothing value must be less than 1.")
        self.smoothing = smoothing
        self.confidence = 1.0 - smoothing

    def forward(self, logits: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward method.

        Args:
            logits: Raw logits from the net.
            target: The targets.

        Returns:
            The computed loss value.
        """
        logprobs = F.log_softmax(logits, dim=-1)
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -logprobs.mean(dim=-1)
        loss = self.confidence * nll_loss + self.smoothing * smooth_loss
        return loss.mean()


class SymmetricCE(nn.Module):
    """Pytorch Implementation of Symmetric Cross Entropy.

    Paper: https://arxiv.org/abs/1908.06112
    """

    def __init__(self, num_classes, alpha: float = 1.0, beta: float = 1.0):
        """Constructor method for symmetric CE.

        Args:
            alpha: The alpha value for symmetricCE.
            beta: The beta value for symmetricCE.
            num_classes: The number of classes.
        """
        super(SymmetricCE, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.num_classes = num_classes
        self.ce = nn.CrossEntropyLoss()

    def forward(self, logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Forward method."""
        ce = self.ce(logits, targets)

        logits = F.softmax(logits, dim=1)
        logits = torch.clamp(logits, min=1e-7, max=1.0)
        if logits.is_cuda:
            label_one_hot = torch.nn.functional.one_hot(targets, self.num_classes).float().cuda()
        else:
            label_one_hot = torch.nn.functional.one_hot(targets, self.num_classes)
        label_one_hot = torch.clamp(label_one_hot, min=1e-4, max=1.0)
        rce = -1 * torch.sum(logits * torch.log(label_one_hot), dim=1)

        loss = self.alpha * ce + self.beta * rce.mean()
        return loss


__all__ = ["LabelSmoothingCrossEntropy", "SymmetricCE"]



"""Implements Load checkpoint."""
from abc import ABC

import torch

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class LoadCheckpoint(Callbacks, ABC):
    """Class to load checkpoint."""

    def __init__(self):
        """Constructor method for LoadCheckpoint Class."""
        super(LoadCheckpoint, self).__init__(order=CallbackOrder.INTERNAL)

    def experiment_start(self):
        """Load checkpoint before starting training."""
        checkpoint = torch.load(self.exp.path, map_location=torch.device(self.exp.device))
        self.exp.model.load_state_dict(checkpoint["model_state_dict"])
        self.exp.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

        if self.exp.scheduler is not None:
            self.exp.scheduler.load_state_dict(checkpoint["scheduler_state_dict"])

        print("Successfully loaded checkpoints.")



"""Implements get optimizer method."""
import torch


def get_optimizer(optimizer):
    """Method to get optimizer from pytorch."""
    dir_optim = dir(torch.optim)
    opts = [o.lower() for o in dir_optim]

    if isinstance(optimizer, str):

        try:
            str_idx = opts.index(optimizer.lower())
            return getattr(torch.optim, dir_optim[str_idx])
        except ValueError:
            raise ValueError("Invalid optimizer string input - must match pytorch optimizer in torch.optim")

    elif hasattr(optimizer, "step") and hasattr(optimizer, "zero_grad"):

        return optimizer

    else:

        raise ValueError("Invalid optimizer input")



"""Implements function for seeding."""
import os
import random

import numpy as np
import torch


def seed_all(seed: int = 42):
    """Method to seed the experiment.

    Args:
        seed : The value for the seed.
    """
    os.environ["PYTHONHASHSEED"] = str(seed)  # set PYTHONHASHSEED env var at fixed value
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)  # pytorch (both CPU and CUDA)
    np.random.seed(seed)  # for numpy pseudo-random generator
    random.seed(seed)  # set fixed value for python built-in pseudo-random generator
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.enabled = False



"""Implements scheduler utilities."""
import torch
import transformers

# Structure of dictionary is as follows:
# 'key' -> scheduler , 'step_on_batch' -> whether to step after batch.
# Set True is step is after batch.
scheduler_step = {
    "LambdaLR": False,
    "MultiplicativeLR": False,
    "StepLR": False,
    "MultiStepLR": False,
    "ExponentialLR": False,
    "CosineAnnealingLR": True,
    "ReduceLROnPlateau": False,
    "CyclicLR": True,
    "OneCycleLR": True,
    "CosineAnnealingWarmRestarts": True,
    "get_constant_schedule": True,
    "get_constant_schedule_with_warmup": True,
    "get_cosine_schedule_with_warmup": True,
    "get_cosine_with_hard_restarts_schedule_with_warmup": True,
    "get_linear_schedule_with_warmup": True,
    "get_polynomial_decay_schedule_with_warmup": True,
}


def get_scheduler(scheduler):
    """Method to get scheduler from pytorch/transformers.

    Args:
        scheduler: The scheduler to be used.

    Returns:
        scheduler.

    Raises:
        ValueError: If scheduler is not found raises value error.
    """
    if isinstance(scheduler, str):

        try:
            if scheduler.startswith("get_"):
                sch = getattr(transformers, scheduler.lower())
            else:
                dir_sch = dir(torch.optim.lr_scheduler)
                opts = [o.lower() for o in dir_sch]
                str_idx = opts.index(scheduler.lower())
                sch = getattr(torch.optim.lr_scheduler, dir_sch[str_idx])

            return sch

        except ValueError:
            raise ValueError(
                "Invalid scheduler string input, must match schedulers available in pytorch or transformers"
            )

    elif hasattr(scheduler, "step"):

        return scheduler

    else:

        raise ValueError("Invalid scheduler input")


class LRScheduler:
    """Class around standard scheduler to decide when to step."""

    def __init__(self, scheduler, **kwargs):
        """Constructor method.

        Args:
            scheduler : The scheduler.
            **kwargs: named arguments for a scheduler.
        """
        self.scheduler = get_scheduler(scheduler)(**kwargs)
        self.step_on_batch = scheduler_step[scheduler]
        self.exp = None

    def set_experiment(self, exp):  # noqa

        self.exp = exp

    def _scheduler_step(self):

        if isinstance(self.scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            val = self.exp.exp_logs.get(self.exp.val_key + self.exp.main_metic)
            self.scheduler.step(val)

        else:
            self.scheduler.step()

    def step(self, current_state):
        """Method to perform the scheduler step.

        Args:
            current_state: The current state of experiment.
        """
        if self.step_on_batch and "batch" in current_state.value:
            self._scheduler_step()

        elif self.step_on_batch is False and "epoch" in current_state.value:
            self._scheduler_step()



"""Implements CosFace."""
import torch
import torch.nn as nn
import torch.nn.functional as F


class CosFace(nn.Module):
    """Implementation of CosFace.

    CosFace: [Large Margin Cosine Loss for Deep Face Recognition](https://arxiv.org/abs/1801.09414)
    """

    def __init__(self, in_features, out_features, s=30.0, m=0.35):
        """Class Constructor.

        Args:
            in_features: Size of the input features
            out_features: The size of output features(usually number of num_classes)
            s: The norm for input features.
            m: margin
        """
        super(CosFace, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.s = s
        self.m = m
        self.Weight = nn.Parameter(torch.FloatTensor(self.out_features, self.in_features))
        nn.init.xavier_uniform_(self.Weight)

    def forward(self, features: torch.Tensor, targets: torch.Tensor = None) -> torch.Tensor:
        """Forward Pass.

        Args:
            features: The input features of shape (BS x F) where BS is batch size and F is input feature dimension.
            targets: The targets with shape BS , where BS is batch size

        Returns:
            Logits with shape (BS x out_features)
        """
        # normalize features and weights
        logits = F.linear(F.normalize(features), F.normalize(self.Weight))
        if targets is None:
            return logits
        # add margin
        target_logits = logits - self.m
        one_hot = torch.zeros_like(logits)
        one_hot.scatter_(1, targets.view(-1, 1).long(), 1)
        output = logits * (1 - one_hot) + target_logits * one_hot
        # feature re-scale
        output *= self.s

        return output


__all__ = ["CosFace"]



"""Implements Model Checkpoint Callback."""
from abc import ABC
from typing import Dict

import numpy as np
import torch

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class ModelCheckpoint(Callbacks, ABC):
    """Callback for Checkpointing your model."""

    def __init__(self, mode: str = "min", monitor: str = "val_loss"):
        """Constructor for ModelCheckpoint class.

        Args:
            mode: One of {"min", "max"}.
                In min mode, training will stop when the quantity monitored has stopped decreasing
                in "max" mode it will stop when the quantity monitored has stopped increasing.
            monitor: The quantity to be monitored. (Default : val_loss)
                    If you want to monitor other metric just pass in the name of the metric.

        Note:

             ModelCheckpoint will save state_dictionaries for model , optimizer , scheduler
             and the value of epoch with following key values:

            1) 'model_state_dict' : The state dictionary of model
            2) 'optimizer_state_dict'  : The state dictionary of optimizer
            3) 'scheduler_state_dict' : The state dictionary of scheduler.(If used)
            4) 'Epoch' : The epoch at which model was saved.

            Model checkpoint will be saved based on the values of metrics/loss obtained from validation set.
        """
        super(ModelCheckpoint, self).__init__(order=CallbackOrder.INTERNAL)
        self.mode = mode
        self.eps = 1e-7
        if "val_" not in monitor:
            self.monitor = "val_" + monitor
        else:
            self.monitor = monitor

        if self.mode == "max":
            self.best_val = -np.inf
            self.improvement = lambda val, best_val: val >= best_val + self.eps
        else:
            self.best_val = np.inf
            self.improvement = lambda val, best_val: val <= best_val + self.eps

    def checkpoint(self, epoch: int):
        """Method to save the state dictionaries of model, optimizer,etc.

        Args:
            epoch : The epoch at which model is saved.
        """
        if self.exp.scheduler_stepper is not None:
            torch.save(
                {
                    "model_state_dict": self.exp.model.state_dict(),
                    "optimizer_state_dict": self.exp.optimizer.state_dict(),
                    "scheduler_state_dict": self.exp.scheduler_stepper.scheduler.state_dict(),
                    "Epoch": epoch,
                },
                self.exp.path,
            )
        else:
            torch.save(
                {
                    "model_state_dict": self.exp.model.state_dict(),
                    "optimizer_state_dict": self.exp.optimizer.state_dict(),
                    "Epoch": epoch,
                },
                self.exp.path,
            )

    def epoch_end(self, epoch: int, logs: Dict):
        """Method to save best model depending on the monitored quantity.

        Args:
            epoch: The current epoch.
            logs: A dictionary containing metrics and loss values.
        """
        val = logs.get(self.monitor)

        if self.improvement(val=val, best_val=self.best_val):

            self.checkpoint(epoch=epoch)



"""Implements Neptune Logger."""
from abc import ABC
from typing import List

import neptune.new as neptune

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class NeptuneLogger(Callbacks, ABC):
    """Callback to log your metrics and loss values to Neptune to track your experiments.

    For more information about Neptune take a look at  [Neptune](https://neptune.ai/)
    """

    def __init__(
        self,
        project_dir: str,
        api_token: str,
        params: dict = None,
        experiment_name: str = None,
        tags: List[str] = None,
    ):
        """Constructor for NeptuneLogger Class.

        Args:
            project_dir: The qualified name of a project in a form of namespace/project_name
            params: he hyperparameters for your model and experiment as a dictionary
            experiment_name: The name of the experiment
            api_token: User’s API token
            tags:  List of strings.
        """
        super(NeptuneLogger, self).__init__(order=CallbackOrder.LOGGING)
        self.project_dir = project_dir
        self.api_token = api_token
        self.params = params
        self.tags = tags
        self.experiment_name = experiment_name

        self.experiment = neptune.init(
            project=self.project_dir, api_token=self.api_token, tags=self.tags, name=self.experiment_name
        )
        self.experiment["params"] = self.params

    def _log_metrics(self, name, value, epoch):

        self.experiment[name].log(value=value, step=epoch)

    def epoch_end(self, epoch, logs):
        """Method to log metrics and values at the end of very epoch.

        Args:
            logs : A dictionary containing metrics and loss values.
            epoch: The current epoch
        """
        for key, value in logs.items():
            if not isinstance(value, str):
                self._log_metrics(name=key, value=value, epoch=epoch)

    def experiment_end(self):
        """Method to end experiment after training is done."""
        self.experiment.stop()



"""Implementation of Callbacks and CallbackRunner."""
from enum import Enum
from typing import Dict, List

from torchflare.callbacks.states import ExperimentStates


def sort_callbacks(callbacks: List) -> List:
    """Method to sort callbacks.

    Args:
        callbacks: List of callbacks.

    Returns:
        Callbacks sorted according to order.
    """
    callbacks = sorted(callbacks, key=lambda cb: cb.order.value)
    return callbacks


class Callbacks:
    """Simple class for defining callbacks depending on the experiment state."""

    def __init__(self, order):
        """Constructor for class callbacks.

        Args:
            order: The priority value for callbacks so that they can be sorted according to the value.
        """
        self.exp = None
        self.order = order

    def set_experiment(self, exp):  # noqa
        self.exp = exp

    def epoch_start(self, epoch, logs):
        """Start of Epoch."""
        raise NotImplementedError

    def epoch_end(self, epoch, logs):
        """End of epoch."""
        raise NotImplementedError

    def experiment_start(self):
        """Start of experiment."""
        raise NotImplementedError

    def experiment_end(self):
        """End of experiment."""
        raise NotImplementedError


class CallbackRunner:
    """Class to run all the callbacks."""

    def __init__(self, callbacks):
        """Constructor for CallbackRunner Class.

        Args:
            callbacks: The List of callbacks
        """
        self.callbacks = callbacks

    def set_experiment(self, exp):  # noqa
        for cb in self.callbacks:
            cb.set_experiment(exp)

    def __call__(self, current_state: Enum, epoch: int = None, logs: Dict = None):
        """Runs callbacks depending on the current experiment state.

        Args:
            current_state : The current model state while training
            logs : A dict containing all the metrics and loss values
            epoch: The current epoch.
        """
        for cb in self.callbacks:
            if current_state.value in list(cb.__class__.__dict__):
                if current_state in (ExperimentStates.EXP_END, ExperimentStates.EXP_START):
                    _ = getattr(cb, current_state.value)()
                else:
                    _ = getattr(cb, current_state.value)(epoch=epoch, logs=logs)



"""Utils for criterion."""
import torch
import torch.nn.functional as F


def normalize(x, axis=-1):
    """Performs L2-Norm."""
    num = x
    denom = torch.norm(x, 2, axis, keepdim=True).expand_as(x) + 1e-12
    return num / denom


# Source : https://github.com/earhian/Humpback-Whale-Identification-1st-/blob/master/models/triplet_loss.py
def euclidean_dist(x, y):
    """Computes Euclidean distance."""
    m, n = x.size(0), y.size(0)
    xx = torch.pow(x, 2).sum(1, keepdim=True).expand(m, n)
    yy = torch.pow(x, 2).sum(1, keepdim=True).expand(m, m).t()
    dist = xx + yy - 2 * torch.matmul(x, y.t())

    dist = dist.clamp(min=1e-12).sqrt()

    return dist


def cosine_dist(x, y):
    """Computes Cosine Distance."""
    x = F.normalize(x, dim=1)
    y = F.normalize(y, dim=1)
    dist = 2 - 2 * torch.mm(x, y.t())
    return dist



"""Implements Experiment class."""
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from torchflare.callbacks.states import ExperimentStates
from torchflare.experiments.simple_utils import to_device
from torchflare.experiments.state import ExperimentState


class Experiment(ExperimentState):
    """Simple Experiment for handling boilerplate code for training, validation and Inference."""

    def __init__(
        self,
        num_epochs: int,
        save_dir: str = "./exp_outputs",
        model_name: str = "model.bin",
        fp16: bool = False,
        device: str = "cuda",
        compute_train_metrics: bool = False,
        using_batch_mixers: bool = False,
        seed: int = 42,
    ):
        """Init method to set up important variables for training and validation.

        Args:
            num_epochs : The number of epochs to save model.
            save_dir : The directory where to save the model, and the log files.
            model_name : The name of '.bin' file.
                Defaults to 'model.bin'
            fp16 : Set this to True if you want to use mixed precision training(Default : False)
            device : The device where you want train your model.
            compute_train_metrics: Whether to compute metrics on training data as well
            using_batch_mixers : Whether using special batch_mixers like cutmix or mixup.
            seed: The seed to ensure reproducibility.

        Note:
            If batch_mixers are used then set compute_train_metrics to False.
            Also, only validation metrics will be computed if  batch_mixers are used.
        """
        super(Experiment, self).__init__(
            num_epochs=num_epochs,
            save_dir=save_dir,
            model_name=model_name,
            fp16=fp16,
            device=device,
            compute_train_metrics=compute_train_metrics,
            using_batch_mixers=using_batch_mixers,
            seed=seed,
        )

    def compile_experiment(
        self,
        model: nn.Module,
        optimizer: Union[torch.optim.Optimizer, str, Any],
        optimizer_params: Dict[str, Union[int, float]],
        criterion: Union[Callable[[torch.Tensor], torch.Tensor], str],
        scheduler: Optional[str] = None,
        scheduler_params: Dict[str, Union[int, float, str]] = None,
        callbacks: List = None,
        metrics: List = None,
        main_metric: Optional[str] = None,
        resume_checkpoint: bool = False,
    ):
        """Configures the model for training and validation.

        Args:
            model: The model to be trained.
            optimizer: The optimizer to be used or name of optimizer.
                        If you pass in the name of the optimizer, only optimizers available in pytorch are supported.
            optimizer_params: The parameters to be used for the optimizer.
            scheduler: The scheduler or the name of the scheduler.
                    If you pass in the name of the scheduler, only scheduler available in
                    pytorch/transformers can be supported.
            scheduler_params: The parameters to be used for the scheduler.
            criterion: The loss function to optimize or name of the loss function.
                    If you pass in the name of the loss function,
                    only loss functions available in pytorch can be supported.
            callbacks: The list of callbacks to be used.
            metrics: The list of metrics to be used.
            main_metric: The name of main metric to be monitored. Use lower case version.
                        For examples , use 'accuracy' instead of 'Accuracy'.
            resume_checkpoint: Whether to resume training from the saved model.

        Note:
            Supports all the schedulers implemented in pytorch/transformers except SWA.
            Support for custom scheduling will be added soon.
        """
        self.model = model
        self.resume_checkpoint = resume_checkpoint
        self.main_metric = main_metric

        self._set_optimizer(optimizer=optimizer, optimizer_params=optimizer_params)
        self._set_scheduler(scheduler=scheduler, scheduler_params=scheduler_params)
        self._set_metrics(metrics=metrics)
        self._set_callbacks(callbacks=callbacks)
        self._set_criterion(criterion=criterion)

    @property
    def update_train_monitor(self):
        """Returns the train monitor dictionary.

        Returns:
            Training monitor dictionary.
        """
        return self._train_monitor

    @update_train_monitor.setter
    def update_train_monitor(self, metrics: Dict):
        self._train_monitor.update(metrics)

    @property
    def update_val_monitor(self):
        """Returns the validation  monitor dictionary.

        Returns:
            Validation monitor dictionary.
        """
        return self._val_monitor

    @update_val_monitor.setter
    def update_val_monitor(self, metrics: Dict):
        self._val_monitor.update(metrics)

    def _update_mbar(self, epoch):
        logs = [epoch] + list(self.exp_logs.values())
        self._write_stdout(stats=logs)

    def _process_inputs(self, *args):
        args = to_device(args, self.device)
        return args

    def process_inputs(self, x, y=None):
        """Method to move the inputs and targets to the respective device.

        Args:
            x: The input to the model.
            y: The targets. Defaults to None.

        Returns:
            The input and targets if targets are present else returns inputs
        """
        if y is not None:
            x, y = self._process_inputs(x, y)
            return x, y

        else:
            x = self._process_inputs(x)
            return x[0] if len(x) == 1 else x

    def _update_model_logs(self, epoch: int = None):

        # To-do : Better logs updating
        self.exp_logs.update({"Epoch": epoch, **self._train_monitor, **self._val_monitor})

    def _compute_loss(self, op: torch.Tensor, y: Union[torch.Tensor, List[torch.Tensor]]) -> torch.Tensor:
        """Computes loss given the inputs and targets.

        Args:
            op: The output of the net
            y: The targets

        Returns:
            Computed loss
        """
        if isinstance(op, (list, tuple)):
            vals = [self.criterion(ele, y) for ele in op]
            loss = sum(vals)
        else:
            loss = self.criterion(op, y)

        return loss

    def _model_forward_pass(self, x):

        if isinstance(x, (list, tuple)):
            op = self.model(*x)
        elif isinstance(x, dict):
            op = self.model(**x)
        else:
            op = self.model(x)

        return op

    def _calculate_loss(self, x: torch.Tensor, y: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Function to calculate loss and update metric states.

        Args:
            x : The input to the network
            y: The targets.

        Returns:
              The computed loss and output
        """
        x, y = self.process_inputs(x, y)
        op = self._model_forward_pass(x=x)
        loss = self._compute_loss(op=op, y=y)

        return loss, op

    def run_batches(self, iterator, prefix: str, func: Callable):
        """Function to iterate the dataloader through all the batches.

        Args:
            iterator: The dataloader or any iterator which will yield the inputs and targets.
            prefix: The prefix for training/validation. .
            func: The function which will compute the loss , outputs and return them.

        Returns:
            A dictionary containing metrics and loss.
        """
        # create progress bar and set compute_flag
        self._before_step(iterator=iterator, prefix=prefix)

        # reset metrics
        self._metric_runner.reset()

        for x, y in self.progress_bar:
            self.set_state = ExperimentStates.BATCH_START

            loss, op = func(x, y)

            # accumulate values for metric computation
            self._metric_runner.accumulate(op=op, y=y, loss=loss.item(), n=iterator.batch_size)
            self._update_pbar(prefix=prefix, val=loss.item())

            self.set_state = ExperimentStates.BATCH_END

        # compute metrics
        metrics = self._metric_runner.compute(prefix=prefix)
        return metrics

    def fp16_step(self, inputs: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Method to perform mixed precision type update.

        Args:
            inputs : The inputs obtained from the dataloader
            targets: The targets obtained from the dataloader

        Returns:
            The computed loss and outputs
        """
        self.optimizer.zero_grad()

        with torch.cuda.amp.autocast():
            loss, op = self._calculate_loss(x=inputs, y=targets)

        self.scaler.scale(loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()

        return loss, op

    def standard_step(self, inputs: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Method to perform the standard update.

        Args:
            inputs : The inputs obtained from the dataloader
            targets: The targets obtained from the dataloader

        Returns:
            The computed loss and outputs
        """
        self.optimizer.zero_grad()
        loss, op = self._calculate_loss(x=inputs, y=targets)
        loss.backward()
        self.optimizer.step()

        return loss, op

    def train_step(self, inputs: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Method to perform the train step and step scheduler.

        Args:
            inputs: The input to the model.
            targets: The targets.

        Returns:
            Return loss and model output.
        """
        func = "fp16_step" if self.fp16 else "standard_step"
        loss, op = getattr(self, func)(inputs=inputs, targets=targets)
        if self.scheduler_stepper is not None:
            self.scheduler_stepper.step(current_state=self.experiment_state)
        return loss, op

    def val_step(self, inputs: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Method to perform validation step.

        Args:
            inputs: The input to the model.
            targets: The targets.

        Returns:
            Return loss and model outputs.
        """
        with torch.no_grad():
            loss, op = self._calculate_loss(x=inputs, y=targets)
        return loss, op

    def _do_train_epoch(self):
        """Method to train the model for one epoch."""
        self.model.train()
        metrics = self.run_batches(iterator=self._train_dl, prefix=self.train_key, func=self.train_step)

        # GENERATE AND UPDATE METRIC MONITORS
        self.update_train_monitor = metrics

    def _do_val_epoch(self):
        """Method to validate model for one epoch."""
        self.model.eval()
        metrics = self.run_batches(iterator=self._valid_dl, prefix=self.val_key, func=self.val_step)

        # GENERATE AND UPDATE THE MONITORS
        self.update_val_monitor = metrics

    def _do_epoch(self, current_epoch):
        self._do_train_epoch()
        self._do_val_epoch()
        self._update_model_logs(epoch=current_epoch)
        if self.scheduler_stepper is not None:
            self.scheduler_stepper.step(current_state=self.experiment_state)

    def _run(self):
        """Method to run experiment for full number of epochs."""
        for epoch in self.master_bar:

            self.set_state = ExperimentStates.EPOCH_START
            self._do_epoch(current_epoch=epoch)

            # RUNNING CALLBACKS AFTER EVERY EPOCH IS DONE
            self.set_state = ExperimentStates.EPOCH_END
            self._update_mbar(epoch=epoch)

            if self.stop_training:
                break

    def run_experiment(
        self, train_dl: DataLoader, valid_dl: DataLoader,
    ):
        """Method to train and validate the model for fixed number of epochs.

        Args:
            train_dl : The training dataloader.
            valid_dl : The validation dataloader.

        Note:
            Model will only be saved when ModelCheckpoint callback is used.
        """
        # Seed and move to model to device
        self._run_event(event="initialize", train_dl=train_dl, valid_dl=valid_dl)
        # Set the state to TRAIN START
        self.set_state = ExperimentStates.EXP_START
        self._run()
        # Set the state to TRAIN END and run callbacks to ensure everything is completed
        self.set_state = ExperimentStates.EXP_END
        # Perform cleanup , set monitors and logs to empty dicts
        self._run_event(event="cleanup")

    @torch.no_grad()
    def _infer_on_batch(self, inp):

        inp = self.process_inputs(x=inp, y=None)
        op = self._model_forward_pass(x=inp)

        return op.detach().cpu()

    @torch.no_grad()
    def infer(self, test_loader: torch.utils.data.DataLoader, path: str, device: str = "cuda") -> torch.Tensor:
        """Method to perform inference on test dataloader.

        Args:
            test_loader: The dataloader to be use for testing.
            device: The device on which you want to perform inference.
            path: The path where the '.bin' or '.pt' or '.pth' file is saved.
                example: path = "/output/model.bin"

        Yields:
            Output per batch
        """
        # move model to device
        self._model_to_device()
        ckpt = torch.load(path, map_location=torch.device(device))
        self.model.load_state_dict(ckpt["model_state_dict"])

        for inp in test_loader:
            op = self._infer_on_batch(inp=inp)
            yield op

    # Perform a sanity check for the forward pass of them model. Ensuring model is defined correctly.
    def perform_sanity_check(self, dl: DataLoader):
        """Method to check if the model forward pass and loss_computation is working or not.

        Args:
            dl: A PyTorch dataloader.
        """
        self._model_to_device()
        x, y = next(iter(dl))
        x, y = self.process_inputs(x=x, y=y)
        op = self._model_forward_pass(x=x)
        loss = self.criterion(op, y)
        print("Sanity Check Completed. Model Forward Pass and Loss Computation Successful")
        print(f"Output Shape : {op.shape}")
        print(f"Loss for a batch :{loss.item()}")



"""Implements TimeCallback."""
import time

from fastprogress.fastprogress import format_time

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class TimeCallback(Callbacks):
    """Callback to measure time per epoch."""

    def __init__(self):
        """Constructor method for TimeCallback."""
        super(TimeCallback, self).__init__(order=CallbackOrder.INTERNAL)
        self.time = None

    def epoch_start(self, epoch, logs):
        """Method ot start measuring time at the start of epoch."""
        self.time = time.time()

    def epoch_end(self, epoch, logs):
        """Method to stop measuring time at the end of epoch and format it."""
        end = time.time() - self.time
        end = format_time(end)
        self.exp.exp_logs.update({"Time": end})


__all__ = ["TimeCallback"]



"""Implements IouLoss."""
import torch
import torch.nn as nn

from torchflare.metrics.iou_meter import IOU


class IOULoss(nn.Module):
    """Computes intersection over union Loss.

    IOULoss =  1 - iou_score
    """

    def __init__(self, class_dim=1):
        """Constructor method for IOULoss.

        Args:
            class_dim: indicates class dimension (K) for
                outputs and targets tensors (default = 1)
        """
        super(IOULoss, self).__init__()
        self.iou = IOU(threshold=None, class_dim=class_dim)

    def forward(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Forward Method.

        Args:
            outputs: outputs from the net after applying activations.
            targets: The targets.

        Returns:
            The computed loss value.
        """
        self.iou.reset()
        self.iou.accumulate(outputs=outputs, targets=targets)
        return 1 - self.iou.compute()


__all__ = ["IOULoss"]



"""Setup.py for torchflare."""
# flake8: noqa
import os

from setuptools import find_packages, setup

NAME = "torchflare"
URL = "https://github.com/Atharva-Phatak/torchflare"
EMAIL = "athp456@gmail.com"
AUTHOR = "Atharva Phatak"
REQUIRES_PYTHON = ">=3.7.0"
DESCRIPTION = (
    "TorchFlare is a simple, beginner-friendly, and easy-to-use PyTorch Framework train your models effortlessly."
)

current_file_path = os.path.abspath(os.path.dirname(__file__))


readme_file_path = os.path.join(current_file_path, "README.md")
with open(readme_file_path, "r", encoding="utf-8") as f:
    readme = f.read()


version_file_path = os.path.join(current_file_path, "version.txt")
with open(version_file_path, "r", encoding="utf-8") as f:
    version = f.read().strip()

with open(os.path.join(current_file_path, "requirements.txt"), "r") as f:
    requirements = f.read().splitlines()


setup(
    name=NAME,
    version=version,
    description=DESCRIPTION,
    long_description=readme,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    download_url=URL,
    python_requires=REQUIRES_PYTHON,
    project_urls={
        "Bug Tracker": "https://github.com/Atharva-Phatak/torchflare/issues",
        "Documentation": "https://atharva-phatak.github.io/torchflare/",
        "Source Code": "https://github.com/Atharva-Phatak/torchflare",
    },
    install_requires=requirements,
    include_package_data=True,
    zip_safe=False,
    license="Apache License 2.0",
    keywords=["Deep Learning", "Computer Vision", "Natural Language Processing", "PyTorch",],
    packages=find_packages(exclude=["tests", "docs"]),
    classifiers=[
        "Environment :: Console",
        "Natural Language :: English",
        # How mature is this project? Common values are
        #   3 - Alpha, 4 - Beta, 5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        # Pick your license as you wish
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



"""Implementation of utilities for metrics."""
from typing import Tuple

import torch
from einops import reduce


class _BaseMetric:
    """A Class which decides type of classification i.e. binary,multilabel or multiclass."""

    def __init__(self, multilabel: bool = False):
        """Constructor class for BaseMetric class.

        Args:
            multilabel: Set to True if problem type is multilabel.
        """
        self.multilabel = multilabel
        self.case_type = None

    @staticmethod
    def _check_shape(outputs: torch.Tensor, targets: torch.Tensor):
        """Function to check if there is a mismatch between outputs and targets.

        Args:
            outputs: The outputs of the net.
            targets: The targets.

        Raises:
            ValueError:  If shapes does not match.
        """
        if not (outputs.ndim == targets.ndim or outputs.ndim == targets.ndim + 1):
            raise ValueError("Preds and Targets must have same number of dimensions")

    @staticmethod
    def _convert_to_onehot(num_classes: int, indices: torch.Tensor) -> torch.Tensor:
        """Converts tensor to one_hot representation.

        Args:
            num_classes: The number of classes.
            indices: torch.Tensor.

        Returns:
            one_hot converted tensor.
        """
        onehot = torch.zeros(indices.shape[0], num_classes, *indices.shape[1:], dtype=indices.dtype)
        index = indices.long().unsqueeze(1).expand_as(onehot)
        return onehot.scatter_(1, index, 1.0)

    @staticmethod
    def detach_tensor(x: torch.Tensor) -> torch.Tensor:
        """Detaches the tensor."""
        return x.detach().cpu()

    # noinspection PyUnboundLocalVariable
    def _check_type(self, outputs: torch.Tensor, targets: torch.Tensor):
        """Method to infer type of the problem."""
        self._check_shape(outputs, targets)

        if targets.ndim + 1 == outputs.ndim:
            if outputs.shape[1] == 1:
                case_type = "binary"
            else:
                case_type = "multiclass"
        elif outputs.ndim == targets.ndim:
            if self.multilabel:
                case_type = "multilabel"
            else:
                case_type = "binary"

        if self.case_type is None:
            self.case_type = case_type


class _BaseInputHandler(_BaseMetric):
    """Class to handle shapes for various classification tasks."""

    def __init__(
        self, num_classes: int, threshold: float = 0.5, multilabel: bool = False, average: str = "macro",
    ):
        """Constructor method.

        Args:
            num_classes: The number of classes.
            threshold: The threshold for binarization.
            multilabel: Whether the problem is multilabel or not.
            average: One of macro or micro.
        """
        super(_BaseInputHandler, self).__init__(multilabel=multilabel)
        self.num_classes = num_classes
        self.threshold = threshold
        self.multilabel = multilabel
        self.eps = 1e-20
        self.average = average
        assert self.average in ["micro", "macro"], "Average should be one of ['micro , 'macro'] "  # noqa: S101

    @staticmethod
    def _calculate_stats(
        true_preds: torch.Tensor, false_preds: torch.Tensor, pos_preds: torch.Tensor, neg_preds: torch.Tensor,
    ):
        tp = true_preds * pos_preds
        fp = false_preds * pos_preds
        tn = true_preds * neg_preds
        fn = false_preds * neg_preds

        return tp, fp, tn, fn

    def compute_stats(
        self, outputs: torch.Tensor, targets: torch.Tensor,
    ):
        """Computes true_positives, false_positives, true_negatives, false_negatives.

        Args:
            outputs: The outputs of the net.
            targets: The targets.

        Returns:
            True positives , false positives, true negatives , false negatives.
        """
        outputs, targets = self._compute(outputs=outputs, targets=targets)

        true_preds = torch.eq(targets, outputs)
        false_preds = ~true_preds

        pos_preds = torch.eq(outputs, 1.0)
        neg_preds = torch.eq(outputs, 0.0)

        # Some einops operations
        pattern = "r c -> c" if self.average == "macro" else "r c -> "

        tp, fp, tn, fn = self._calculate_stats(true_preds, false_preds, pos_preds, neg_preds)

        # einops reductions
        tp = reduce(tp, pattern, reduction="sum")
        fp = reduce(fp, pattern, reduction="sum")
        tn = reduce(tn, pattern, reduction="sum")
        fn = reduce(fn, pattern, reduction="sum")

        return tp, fp, tn, fn

    def reduce(self, numerator: torch.Tensor, denominator: torch.Tensor) -> torch.Tensor:
        """Method to perform macro or micro reduction."""
        frac = numerator / (denominator + self.eps)
        return torch.mean(frac) if self.average == "macro" else frac

    def _compute(self, outputs: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:

        self._check_type(outputs=outputs, targets=targets)

        if self.case_type == "multiclass":

            targets = self._convert_to_onehot(num_classes=self.num_classes, indices=targets.view(-1))
            # We receive logits  need argmax on preds
            outputs = torch.argmax(outputs, dim=1)
            outputs = self._convert_to_onehot(num_classes=self.num_classes, indices=outputs.view(-1))

        else:

            # Handling multilabel and binary cases

            outputs = torch.sigmoid(outputs)

            outputs = (outputs >= self.threshold).long()

        outputs = outputs.reshape(outputs.shape[0], -1)
        targets = targets.reshape(targets.shape[0], -1)

        return outputs, targets


def calculate_segmentation_statistics(
    outputs: torch.Tensor, targets: torch.Tensor, class_dim: int = 1, threshold=None
):
    """Compute calculate segmentation statistics.

    Args:
        outputs: torch.Tensor.
        targets: torch.Tensor.
        threshold: threshold for binarization of predictions.
        class_dim: indicates class dimension (K).

    Returns:
        True positives , false positives , false negatives for segmentation task.
    """
    num_dims = len(outputs.shape)

    assert num_dims > 2, "Found only two dimensions, shape should be [bs , C , ...]"  # noqa: S101

    assert outputs.shape == targets.shape, "shape mismatch"  # noqa: S101

    if threshold is not None:
        outputs = (outputs > threshold).float()

    dims = [dim for dim in range(num_dims) if dim != class_dim]

    true_positives = torch.sum(outputs * targets, dim=dims)
    false_positives = torch.sum(outputs * (1 - targets), dim=dims)
    false_negatives = torch.sum(targets * (1 - outputs), dim=dims)

    return true_positives, false_positives, false_negatives


class MetricMeter:
    """Base Class to structuring your metrics."""

    def accumulate(self, outputs, targets):
        """Method to accumulate outputs and targets per the batch."""
        raise NotImplementedError

    def compute(self):
        """Method to compute the metric on epoch end."""
        raise NotImplementedError

    def reset(self):
        """Method to reset the accumulation lists."""
        raise NotImplementedError


__all__ = [
    "_BaseMetric",
    "_BaseInputHandler",
    "MetricMeter",
    "calculate_segmentation_statistics",
]



"""Implementations of mixup , cutmix."""
from typing import Callable, Tuple

import numpy as np
import torch


def mixup(batch: Tuple[torch.Tensor, torch.Tensor], alpha: float = 1.0) -> Tuple:
    """Function to mixup data.

    Mixup: <https://arxiv.org/abs/1710.09412>

    Args:
        batch : Tuple containing the data and targets.
        alpha : beta distribution a=b parameters.

    Returns:
        The mixed image and targets.
    """
    data, targets = batch

    lam = np.random.beta(alpha, alpha) if alpha > 0 else 1
    indices = torch.randperm(data.shape[0])
    mixed_data = lam * data + (1 - lam) * data[indices, :]
    target_a, target_b = targets, targets[indices]

    targets = (target_a, target_b, lam)

    return mixed_data, targets


def random_bbox(data, lam):
    """Function to crop random bboxes.

    Args:
        data: The input data.
        lam: The beta distribution value.

    Returns:
        Co-ordinates of bbox.
    """
    img_h, img_w = data.shape[2:]
    cx = np.random.uniform(0, img_w)
    cy = np.random.uniform(0, img_h)
    w = img_w * np.sqrt(1 - lam)
    h = img_h * np.sqrt(1 - lam)
    x0 = int(np.round(max(cx - w / 2, 0)))
    x1 = int(np.round(min(cx + w / 2, img_w)))
    y0 = int(np.round(max(cy - h / 2, 0)))
    y1 = int(np.round(min(cy + h / 2, img_h)))

    return x0, x1, y0, y1


def cutmix(batch: Tuple[torch.Tensor, torch.Tensor], alpha: float = 1.0) -> Tuple:
    """Function to perform cutmix.

    Cutmix: <https://arxiv.org/abs/1905.04899>

    Args:
        batch : Tuple containing the data and targets.
        alpha : beta distribution a=b parameters.

    Returns:
        Image and targets.
    """
    data, targets = batch
    indices = torch.randperm(data.size(0))
    shuffled_data = data[indices]
    shuffled_targets = targets[indices]
    lam = np.random.beta(alpha, alpha) if alpha > 0 else 1

    x0, x1, y0, y1 = random_bbox(data, lam)

    data[:, :, y0:y1, x0:x1] = shuffled_data[:, :, y0:y1, x0:x1]

    targets = (targets, shuffled_targets, lam)

    return data, targets


class MixCriterion:
    """Class to calculate loss when batch mixers are used."""

    def __init__(self, criterion: Callable):
        """Constructor Class for MixCriterion.

        Args:
            criterion: The criterion to be used.
        """
        self.criterion = criterion

    def __call__(self, preds: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Method to calculate loss.

        Args:
            preds: The output of network.
            targets: The targets.

        Returns:
            The computed loss.
        """
        if isinstance(targets, (list, tuple)):

            target_a, target_b, lam = targets
            loss = lam * self.criterion(preds, target_a) + (1 - lam) * self.criterion(preds, target_b)
        else:
            loss = self.criterion(preds, targets)
        return loss


class CustomCollate:
    """Class to create custom collate_fn for dataloaders."""

    def __init__(self, mixer: Callable, alpha: float = 1.0):
        """Constructor for CustomCollate class.

        Args:
            mixer: The batch mix function to be used.
            alpha: beta distribution a=b parameters.
        """
        self.alpha = alpha
        self.aug = mixer

    def __call__(self, batch):
        """Call method.

        Args:
            batch : The input batch from dataloader.

        Returns:
            Batch with a mixer applied.
        """
        batch = torch.utils.data.dataloader.default_collate(batch)
        batch = self.aug(batch, self.alpha)

        return batch


def get_collate_fn(mixer_name: str, alpha: float) -> Callable:
    """Method to create  collate_fn for dataloader.

    Args:
        mixer_name: The name of the batch_mixer.
        alpha: beta distribution a=b parameters.

    Returns:
        The collate_fn for the respective special augmentation.

    Note:
        aug_name must be one of cutmix , mixup
    """
    fn = cutmix if mixer_name == "cutmix" else mixup
    collate_fn = CustomCollate(alpha=alpha, mixer=fn)
    return collate_fn


__all__ = ["mixup", "cutmix", "CustomCollate", "MixCriterion", "get_collate_fn"]



"""Imports for mixers."""
from torchflare.batch_mixers.mixers import CustomCollate, MixCriterion, cutmix, get_collate_fn, mixup

__all__ = ["CustomCollate", "MixCriterion", "cutmix", "get_collate_fn", "mixup"]



"""Metric imports."""
from torchflare.metrics.accuracy_meter import Accuracy
from torchflare.metrics.dice_meter import DiceScore
from torchflare.metrics.fbeta_meter import F1Score, FBeta
from torchflare.metrics.iou_meter import IOU
from torchflare.metrics.meters import MetricMeter, _BaseInputHandler, _BaseMetric, calculate_segmentation_statistics
from torchflare.metrics.metric_utils import MetricAndLossContainer
from torchflare.metrics.precision_meter import Precision
from torchflare.metrics.recall_meter import Recall
from torchflare.metrics.regression import MAE, MSE, MSLE, R2Score

__all__ = [
    "Accuracy",
    "Precision",
    "Recall",
    "FBeta",
    "F1Score",
    "DiceScore",
    "IOU",
    "_BaseMetric",
    "_BaseInputHandler",
    "MetricMeter",
    "MetricAndLossContainer",
    "R2Score",
    "MSLE",
    "MAE",
    "MSE",
    "calculate_segmentation_statistics",
]



"""Simple utilities required by experiment."""
from typing import List

import numpy as np
import torch


def wrap_metric_names(metric_list: List):
    """Method to  get the metric names from metric list.

    Args:
        metric_list : A list of metrics.

    Returns:
        list of metric name.s
    """
    return [metric.handle() for metric in metric_list]


def to_device(value, device):
    """Move tensor, list of tensors, list of list of tensors, dict of tensors, tuple of tensors to target device.

    Args:
        value: Object to be moved to the device.
        device: target device.

    Returns:
        Same structure as value, but all tensors and np.arrays moved to device
    """
    if isinstance(value, dict):
        return {k: to_device(v, device) for k, v in value.items()}
    elif isinstance(value, (tuple, list)):
        return [to_device(v, device) for v in value]
    elif torch.is_tensor(value):
        return value.to(device, non_blocking=True)
    elif isinstance(value, (np.ndarray, np.void)) and value.dtype.fields is not None:
        return {k: to_device(value[k], device) for k in value.dtype.fields.keys()}
    elif isinstance(value, np.ndarray):
        return torch.tensor(value, device=device)
    return value


def to_numpy(x):
    """Convert tensors to numpy array.

    Args:
        x : The input tensor.

    Returns:
        Numpy array.
    """
    return x.detach().cpu().numpy()



"""Implements Recall-score."""
import torch

from torchflare.metrics.meters import MetricMeter, _BaseInputHandler


class Recall(_BaseInputHandler, MetricMeter):
    """Class to compute Recall Score.

    Support binary, multiclass and multilabel cases
    """

    def __init__(
        self, num_classes: int, average: str = "macro", threshold: float = 0.5, multilabel: bool = False,
    ):
        """Constructor method for Precision Class.

        Args:
            num_classes: The number of num_classes.
            average: The type of reduction to apply.
                macro: calculate metrics for each class and averages them with equal weightage to each class.
                micro: calculate metrics globally for each sample and class.
            threshold: The threshold value to transform probability predictions to binary values(0,1)
            multilabel: Set it to True if your problem is  multilabel classification.
        """
        super(Recall, self).__init__(
            num_classes=num_classes, threshold=threshold, multilabel=multilabel, average=average,
        )

        self._outputs = None
        self._targets = None

        self.reset()

    def handle(self) -> str:
        """Method to get the class name.

        Returns:
            The name of the class
        """
        return self.__class__.__name__.lower()

    def accumulate(self, outputs: torch.Tensor, targets: torch.Tensor):
        """Accumulates the batch outputs and targets.

        Args:
            outputs : raw logits from the network.
            targets : targets to use for computing accuracy
        """
        outputs, targets = self.detach_tensor(outputs), self.detach_tensor(targets)
        self._outputs.append(outputs)
        self._targets.append(targets)

    def compute(self) -> torch.Tensor:
        """Compute the recall score.

        Returns:
            The computed recall score.
        """
        outputs = torch.cat(self._outputs)
        targets = torch.cat(self._targets)

        tp, fp, tn, fn = self.compute_stats(outputs=outputs, targets=targets)
        recall = self.reduce(numerator=tp, denominator=tp + fn)
        return recall

    def reset(self):
        """Reset the output and target lists."""
        self._outputs = []
        self._targets = []


__all__ = ["Recall"]



"""Implements datasets for task involving tabular data."""
from __future__ import annotations

from typing import List, Union

import pandas as pd
import torch
from torch.utils.data import Dataset


class TabularDataset(Dataset):
    """Class to create a dataset for tasks involving tabular data."""

    def __init__(self, features, labels=None):
        """Constructor for Tabular Dataset class.

        Args:
            features: A np.array of features.
            labels : A np.array of labels.
        """
        self.features = features
        self.labels = labels

    @classmethod
    def from_df(
        cls, df: pd.DataFrame, feature_cols: Union[str, List[str]], label_cols: Union[str, List[str]] = None,
    ) -> TabularDataset:
        """Classmethod to create pytorch style dataset from dataframes.

        Args:
            df: The dataframe which has features, and the labels/targets.
            feature_cols: The name of columns which contain features.
                feature_cols can be a string if single column or can be a list of string if multiple columns.
            label_cols: The name of columns which contain the labels.
                label_cols can be a string or can be a list of string if multiple columns are used.

        Returns:
            np.array of features , labels if label_cols is not None else return np.array of features.
        """
        features = df.loc[:, feature_cols].values
        labels = df.loc[:, label_cols].values if label_cols is not None else None

        return cls(features=features, labels=labels)

    @classmethod
    def from_csv(
        cls, csv_path, feature_cols: Union[str, List[str]], label_cols: Union[str, List[str]] = None,
    ) -> TabularDataset:
        """Classmethod to create pytorch style dataset from csv file.

        Args:
            csv_path: The full path to csv.
            feature_cols: The name of columns which contain features.
                feature_cols can be a string if single column or can be a list of string if multiple columns.
            label_cols: The name of columns which contain the labels.
                label_cols can be a string or can be a list of string if multiple columns are used.

        Returns:
            np.array of features , labels if label_cols is not None else return np.array of features.
        """
        return cls.from_df(df=pd.read_csv(csv_path), feature_cols=feature_cols, label_cols=label_cols)

    def __len__(self):
        """__len__ method.

        Returns:
            The length of dataloader.
        """
        return self.features.shape[0]

    def __getitem__(self, item):
        """__getitem__ method.

        Args:
            item: idx

        Returns:
            Tensors of features , labels if labels is not None else returns Tensors of features.
        """
        feature = torch.tensor(self.features[item], dtype=torch.long)
        if self.labels is not None:
            labels = torch.tensor(self.labels[item], dtype=torch.long)
            return feature, labels

        return feature



"""Implements variants for Focal loss."""
import torch
import torch.nn as nn
import torch.nn.functional as F


class BCEFocalLoss(nn.Module):
    """Implementation of Focal Loss for Binary Classification Problems.

    Focal loss was proposed in [Focal Loss for Dense Object Detection](https://arxiv.org/abs/1708.02002).
    """

    def __init__(self, gamma=0, eps=1e-7, reduction="mean"):
        """Constructor Method for FocalLoss class.

        Args:
            gamma : The focal parameter. Defaults to 0.
            eps : Constant for computational stability.
            reduction: The reduction parameter for Cross Entropy Loss.
        """
        super(BCEFocalLoss, self).__init__()
        self.gamma = gamma
        self.reduction = reduction
        self.eps = eps
        self.bce = torch.nn.BCEWithLogitsLoss(reduction="none")

    def forward(self, logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Forward method.

        Args:
            logits: The raw logits from the network of shape (N,*) where C = number of classes , * = extra dims
            targets: The targets

        Returns:
            The computed loss value
        """
        targets = targets.view(logits.shape)
        logp = self.bce(logits, targets)
        p = torch.exp(-logp)
        loss = (1 - p) ** self.gamma * logp
        return loss.mean() if self.reduction == "mean" else loss.sum() if self.reduction == "sum" else loss


class FocalLoss(nn.Module):
    """Implementation of Focal Loss.

    Focal loss was proposed in [Focal Loss for Dense Object Detection](https://arxiv.org/abs/1708.02002).
    """

    def __init__(self, gamma=0, eps=1e-7, reduction="mean"):
        """Constructor Method for FocalLoss class.

        Args:
            gamma : The focal parameter. Defaults to 0.
            eps : Constant for computational stability.
            reduction: The reduction parameter for Cross Entropy Loss.
        """
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.reduction = reduction
        self.eps = eps
        self.ce = torch.nn.CrossEntropyLoss(reduction="none")

    def forward(self, logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Forward method.

        Args:
            logits: The raw logits from the network of shape (N,C,*) where C = number of classes , * = extra dims
            targets: The targets of shape (N , *).

        Returns:
            The computed loss value
        """
        logp = self.ce(logits, targets)
        p = torch.exp(-logp)
        loss = (1 - p) ** self.gamma * logp
        return loss.mean() if self.reduction == "mean" else loss.sum() if self.reduction == "sum" else loss


class FocalCosineLoss(nn.Module):
    """Implementation Focal cosine loss.

    [Data-Efficient Deep Learning Method for Image Classification
    Using Data Augmentation, Focal Cosine Loss, and Ensemble](https://arxiv.org/abs/2007.07805).

    Source : <https://www.kaggle.com/c/cassava-leaf-disease-classification/discussion/203271>
    """

    def __init__(self, alpha: float = 1, gamma: float = 2, xent: float = 0.1, reduction="mean"):
        """Constructor for FocalCosineLoss.

        Todo:
        Add documentation.
        """
        super(FocalCosineLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.xent = xent
        self.reduction = reduction

    def forward(self, logits: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward Method."""
        cosine_loss = F.cosine_embedding_loss(
            logits,
            torch.nn.functional.one_hot(target, num_classes=logits.size(-1)),
            torch.tensor([1], device=target.device),
            reduction=self.reduction,
        )

        cent_loss = F.cross_entropy(F.normalize(logits), target, reduction="none")
        pt = torch.exp(-cent_loss)
        focal_loss = self.alpha * (1 - pt) ** self.gamma * cent_loss

        if self.reduction == "mean":
            focal_loss = torch.mean(focal_loss)

        return cosine_loss + self.xent * focal_loss



"""Implements Comet Logger."""

from abc import ABC
from typing import List

import comet_ml

from torchflare.callbacks.callback import Callbacks
from torchflare.callbacks.states import CallbackOrder


class CometLogger(Callbacks, ABC):
    """Callback to log your metrics and loss values to Comet to track your experiments.

    For more information about Comet look at [Comet.ml](https://www.comet.ml/site/)
    """

    def __init__(
        self, api_token: str, params: dict, project_name: str, workspace: str, tags: List[str],
    ):
        """Constructor for CometLogger class.

        Args:
            api_token: Your API key obtained from comet.ml
            params: The hyperparameters for your model and experiment as a dictionary
            project_name: Send your experiment to a specific project.
                    Otherwise, will be sent to Uncategorized Experiments.
            workspace: Attach an experiment to a project that belongs to this workspace
            tags: List of strings.
        """
        super(CometLogger, self).__init__(order=CallbackOrder.LOGGING)
        self.api_token = api_token
        self.project_name = project_name
        self.workspace = workspace
        self.params = params
        self.tags = tags

        self.experiment = comet_ml.Experiment(
            project_name=self.project_name,
            api_key=self.api_token,
            workspace=self.workspace,
            log_code=False,
            display_summary_level=0,
        )

        if self.tags is not None:
            self.experiment.add_tags(self.tags)

        if self.params is not None:
            self.experiment.log_parameters(self.params)

    def epoch_end(self, epoch: int, logs: dict):
        """Function to log your metrics and values at the end of very epoch.

        Args:
            logs : A dictionary containing metrics and loss values.
            epoch: The current epoch
        """
        _ = logs.pop("Time")
        self.experiment.log_metrics(logs, step=epoch)

    def experiment_end(self):
        """Function to close the experiment when training ends."""
        self.experiment.end()
